"""
Plugin facility to read+normalize information from different sources to a common format
"""
import warnings

import numpy as np
import h5py
import xarray as xr
from astropy.io import fits

from pathlib import Path
import typing
from typing import Any, Dict
import copy
import arpes.config
import arpes.constants
import os.path

from arpes.load_pxt import read_single_pxt, find_ses_files_associated
from arpes.utilities.dict import case_insensitive_get, rename_dataarray_attrs
from arpes.preparation import replace_coords
from arpes.provenance import provenance_from_file
from arpes.endstations.fits_utils import find_clean_coords
from arpes.endstations.igor_utils import shim_wave_note
from arpes.repair import negate_energy
from xarray.core.dataset import Dataset

__all__ = ('endstation_name_from_alias', 'endstation_from_alias', 'add_endstation', 'load_scan',
           'EndstationBase', 'FITSEndstation', 'HemisphericalEndstation', 'SynchrotronEndstation',
           'SingleFileEndstation', 'load_scan_for_endstation',)

_ENDSTATION_ALIASES = {}


class EndstationBase:
    """
    Implements the core features of ARPES data loading. A thorough documentation
    is available at `the plugin documentation <https://arpes.netlify.com/#/writing-plugins>`_.

    To summarize, a plugin has a few core jobs:

    1. Load data, including collating any data that is in a multi-file format
       This is accomplished with `.load`, which delegates loading `frames` (single files)
       to `load_single_frame`. Frame collation is then performed by `concatenate_frames`.
    2. Loading and attaching metadata.
    3. Normalizing metadata to standardized names. These are documented at the
       `data model documentation <https://arpes.netlify.com/#/spectra>`_.
    4. Ensuring all angles and necessary coordinates are attached to the data.
       Data should permit immediate conversion to angle space after being loaded.

    Plugins are in one-to-many correspondance with the values of the "location" column in
    analysis spreadsheets. This binding is provided by PRINCIPAL_NAME and ALIASES.

    The simplest way to normalize metadata is by renaming keys, but sometimes additional
    work is required. RENAME_KEYS is provided to make this simpler, and is implemented in
    scan post-processessing.
    """
    ALIASES = []
    PRINCIPAL_NAME = None
    ATTR_TRANSFORMS = {}
    MERGE_ATTRS = {}

    # adjust as needed
    CONCAT_COORDS = ['hv', 'chi', 'psi', 'timed_power', 'tilt', 'beta', 'theta']

    # phi because this happens sometimes at BL4 with core level scans
    SUMMABLE_NULL_DIMS = ['phi', 'cycle']

    RENAME_KEYS = {}

    def concatenate_frames(self, frames=typing.List[xr.Dataset], scan_desc: dict = None):
        if not frames:
            raise ValueError('Could not read any frames.')

        if len(frames) == 1:
            return frames[0]

        # determine which axis to stitch them together along, and then do this
        scan_coord = None
        max_different_values = -np.inf
        for possible_scan_coord in self.CONCAT_COORDS:
            coordinates = [f.attrs.get(possible_scan_coord, None) for f in frames]
            n_different_values = len(set(coordinates))
            if n_different_values > max_different_values and None not in coordinates:
                max_different_values = n_different_values
                scan_coord = possible_scan_coord

        assert scan_coord is not None

        for f in frames:
            f.coords[scan_coord] = f.attrs[scan_coord]

        frames.sort(key=lambda x: x.coords[scan_coord])
        return xr.concat(frames, scan_coord)

    def resolve_frame_locations(self, scan_desc: dict = None):
        return []

    def load_single_frame(self, frame_path: str = None, scan_desc: dict = None, **kwargs):
        print(frame_path)
        return xr.Dataset()

    def postprocess(self, frame: xr.Dataset):
        from arpes.utilities import rename_keys

        frame = xr.Dataset({
            k: rename_dataarray_attrs(v, self.RENAME_KEYS) for k, v in frame.data_vars.items()
        }, attrs=rename_keys(frame.attrs, self.RENAME_KEYS))

        sum_dims = []
        for dim in frame.dims:
            if len(frame.coords[dim]) == 1 and dim in self.SUMMABLE_NULL_DIMS:
                sum_dims.append(dim)

        if sum_dims:
            frame = frame.sum(sum_dims, keep_attrs=True)

        return frame

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict = None):
        # attach the 'spectrum_type'
        # TODO move this logic into xarray extensions and customize here
        # only as necessary
        coord_names = tuple(sorted([c for c in data.dims if c != 'cycle']))

        spectrum_type = None
        if any(d in coord_names for d in {'x', 'y', 'z'}):
            coord_names = tuple(c for c in coord_names if c not in {'x', 'y', 'z'})
            spectrum_types = {
                ('eV',): 'spem',
                ('eV', 'phi',): 'ucut',
            }
            spectrum_type = spectrum_types.get(coord_names)
        else:
            spectrum_types = {
                ('eV',): 'xps',
                ('eV', 'phi', 'theta',): 'map',
                ('eV', 'phi', 'psi',): 'map',
                ('beta', 'eV', 'phi',): 'map',
                ('eV', 'hv', 'phi',): 'hv_map',
                ('eV', 'phi'): 'cut',
            }
            spectrum_type = spectrum_types.get(coord_names)

        if 'phi' not in data.coords:
            # XPS
            data.coords['phi'] = 0
            for s in data.S.spectra:
                s.coords['phi'] = 0

        if spectrum_type is not None:
            data.attrs['spectrum_type'] = spectrum_type
            if 'spectrum' in data.data_vars:
                data.spectrum.attrs['spectrum_type'] = spectrum_type

        ls = [data] + data.S.spectra
        for l in ls:
            for k, key_fn in self.ATTR_TRANSFORMS.items():
                if k in l.attrs:
                    transformed = key_fn(l.attrs[k])
                    if isinstance(transformed, dict):
                        l.attrs.update(transformed)
                    else:
                        l.attrs[k] = transformed

        for l in ls:
            for k, v in self.MERGE_ATTRS.items():
                if k not in l.attrs:
                    l.attrs[k] = v

        for l in ls:
            for c in ['x', 'y', 'z', 'theta', 'beta', 'chi', 'hv', 'alpha', 'psi']:
                if c not in l.coords:
                    l.coords[c] = l.attrs[c]

        for l in ls:
            if 'chi' in l.coords and 'chi_offset' not in l.attrs:
                l.attrs['chi_offset'] = l.coords['chi'].item()

        return data

    def load(self, scan_desc: dict = None, **kwargs):
        """
        Loads a scan from a single file or a sequence of files.

        :param scan_desc:
        :param kwargs:
        :return:
        """
        resolved_frame_locations = self.resolve_frame_locations(scan_desc)
        resolved_frame_locations = [f if isinstance(f, str) else str(f) for f in resolved_frame_locations]

        frames = [self.load_single_frame(fpath, scan_desc, **kwargs) for fpath in resolved_frame_locations]
        frames = [self.postprocess(f) for f in frames]
        concatted = self.concatenate_frames(frames, scan_desc)
        concatted = self.postprocess_final(concatted, scan_desc)

        if 'id' in scan_desc:
            concatted.attrs['id'] = scan_desc['id']

        return concatted


class SingleFileEndstation(EndstationBase):
    """
    Abstract endstation which loads data from a single file. This just specializes
    the routine used to determine the location of files on disk.

    Unlike general endstations, if your data comes in a single file you can trust that the
    file given to you in the spreadsheet or direct load calls is all there is.
    """
    def resolve_frame_locations(self, scan_desc: dict = None):
        if scan_desc is None:
            raise ValueError('Must pass dictionary as file scan_desc to all endstation loading code.')

        original_data_loc = scan_desc.get('path', scan_desc.get('file'))
        p = Path(original_data_loc)
        if not p.exists():
            original_data_loc = os.path.join(arpes.config.DATA_PATH, original_data_loc)

        p = Path(original_data_loc)
        return [p]


class SESEndstation(EndstationBase):
    """
    Provides collation and data loading for files produced by Scienta's SESWrapper
    and endstations that use the SESWrapper.

    These files have special frame names, at least at the beamlines Conrad has encountered.
    """
    def resolve_frame_locations(self, scan_desc: dict = None):
        if scan_desc is None:
            raise ValueError('Must pass dictionary as file scan_desc to all endstation loading code.')

        original_data_loc = scan_desc.get('path', scan_desc.get('file'))
        p = Path(original_data_loc)
        if not p.exists():
            original_data_loc = os.path.join(arpes.config.DATA_PATH, original_data_loc)

        p = Path(original_data_loc)
        return find_ses_files_associated(p)

    def load_single_frame(self, frame_path: str=None, scan_desc: dict = None, **kwargs):
        name, ext = os.path.splitext(frame_path)

        if 'nc' in ext:
            # was converted to hdf5/NetCDF format with Conrad's Igor scripts
            scan_desc = copy.deepcopy(scan_desc)
            scan_desc['path'] = frame_path
            return self.load_SES_nc(scan_desc=scan_desc, **kwargs)

        # it's given by SES PXT files
        pxt_data = negate_energy(read_single_pxt(frame_path))
        return xr.Dataset({'spectrum': pxt_data}, attrs=pxt_data.attrs)

    def postprocess(self, frame: xr.Dataset):
        import arpes.xarray_extensions  # pylint: disable=unused-import, redefined-outer-name

        frame = super().postprocess(frame)
        return frame.assign_attrs(frame.S.spectrum.attrs)

    def load_SES_nc(self, scan_desc: dict = None, robust_dimension_labels=False, **kwargs):
        """
        Imports an hdf5 dataset exported from Igor that was originally generated by a Scienta spectrometer
        in the SESb format. In order to understand the structure of these files have a look at Conrad's
        saveSESDataset in Igor Pro.

        :param scan_desc: Dictionary with extra information to attach to the xr.Dataset, must contain the location
        of the file
        :return:
        """

        scan_desc = copy.deepcopy(scan_desc)

        data_loc = scan_desc.get('path', scan_desc.get('file'))
        p = Path(data_loc)
        if not p.exists():
            data_loc = os.path.join(arpes.config.DATA_PATH, data_loc)

        wave_note = shim_wave_note(data_loc)
        f = h5py.File(data_loc, 'r')

        primary_dataset_name = list(f)[0]
        # This is bugged for the moment in h5py due to an inability to read fixed length unicode strings
        # wave_note = f['/' + primary_dataset_name].attrs['IGORWaveNote']

        # Use dimension labels instead of
        dimension_labels = list(f['/' + primary_dataset_name].attrs['IGORWaveDimensionLabels'][0])
        if any(x == '' for x in dimension_labels):
            print(dimension_labels)

            if not robust_dimension_labels:
                raise ValueError('Missing dimension labels. Use robust_dimension_labels=True to override')
            else:
                used_blanks = 0
                for i in range(len(dimension_labels)):
                    if dimension_labels[i] == '':
                        dimension_labels[i] = 'missing{}'.format(used_blanks)
                        used_blanks += 1

                print(dimension_labels)

        scaling = f['/' + primary_dataset_name].attrs['IGORWaveScaling'][-len(dimension_labels):]
        raw_data = f['/' + primary_dataset_name][:]

        scaling = [np.linspace(scale[1], scale[1] + scale[0] * raw_data.shape[i], raw_data.shape[i])
                   for i, scale in enumerate(scaling)]

        dataset_contents = {}
        attrs = scan_desc.pop('note', {})
        attrs.update(wave_note)

        built_coords = dict(zip(dimension_labels, scaling))

        deg_to_rad_coords = {'theta', 'beta', 'phi', 'alpha', 'psi'}

        # the hemisphere axis is handled below
        built_coords = {k: c * (np.pi / 180) if k in deg_to_rad_coords else c
                        for k, c in built_coords.items()}

        deg_to_rad_attrs = {'theta', 'beta', 'alpha', 'psi', 'chi'}
        for angle_attr in deg_to_rad_attrs:
            if angle_attr in attrs:
                attrs[angle_attr] = float(attrs[angle_attr]) * np.pi / 180

        dataset_contents['spectrum'] = xr.DataArray(
            raw_data,
            coords=built_coords,
            dims=dimension_labels,
            attrs=attrs,
        )

        provenance_from_file(dataset_contents['spectrum'], data_loc, {
            'what': 'Loaded SES dataset from HDF5.',
            'by': 'load_SES'
        })

        return xr.Dataset(
            dataset_contents,
            attrs={**scan_desc, 'name': primary_dataset_name},
        )


class FITSEndstation(EndstationBase):
    """
    Loads data from the .fits format produced by the MAESTRO software and derivatives.

    This ends up being somewhat complicated, because the FITS export is written in LabView and
    does not conform to the standard specification for the FITS archive format.

    Many of the intricaces here are in fact those shared between MAESTRO's format
    and the Lanzara Lab's format. Conrad does not forsee this as an issue, because it is
    unlikely that mnay other ARPES labs will adopt this data format moving forward, in
    light of better options derivative of HDF like the NeXuS format.
    """
    PREPPED_COLUMN_NAMES = {
        'time': 'time',
        'Delay': 'delay-var',  # these are named thus to avoid conflicts with the
        'Sample-X': 'cycle-var',  # underlying coordinates
        'Mira': 'pump_power',
        # insert more as needed
    }

    SKIP_COLUMN_NAMES = {
        'Phi',
        'null',
        'X',
        'Y',
        'Z',
        'mono_eV',
        'Slit Defl',
        'Optics Stage',
        'Scan X',
        'Scan Y',
        'Scan Z',
        # insert more as needed
    }

    SKIP_COLUMN_FORMULAS = {
        lambda name: True if ('beamview' in name or 'IMAQdx' in name) else False,
    }

    RENAME_KEYS = {
        'Phi': 'chi',
        'Beta': 'beta',
        'Azimuth': 'chi',
        'Pump_energy_uJcm2': 'pump_fluence',
        'T0_ps': 't0_nominal',
        'W_func': 'workfunction',
        'Slit': 'slit',
        'LMOTOR0': 'x',
        'LMOTOR1': 'y',
        'LMOTOR2': 'z',
        'LMOTOR3': 'theta',
        'LMOTOR4': 'beta',
        'LMOTOR5': 'chi',
        'LMOTOR6': 'alpha',
    }

    def resolve_frame_locations(self, scan_desc: dict = None):
        if scan_desc is None:
            raise ValueError('Must pass dictionary as file scan_desc to all endstation loading code.')

        original_data_loc = scan_desc.get('path', scan_desc.get('file'))
        p = Path(original_data_loc)
        if not p.exists():
            original_data_loc = os.path.join(arpes.config.DATA_PATH, original_data_loc)

        return [original_data_loc]

    def load_single_frame(self, frame_path: str=None, scan_desc: dict = None, **kwargs):
        # Use dimension labels instead of
        hdulist = fits.open(frame_path, ignore_missing_end=True)
        primary_dataset_name = None

        # Clean the header because sometimes out LabView produces improper FITS files
        for i in range(len(hdulist)):
            # This looks a little stupid, but because of confusing astropy internals actually works
            hdulist[i].header['UN_0_0'] = ''  # TODO This card is broken, this is not a good fix
            del hdulist[i].header['UN_0_0']
            hdulist[i].header['UN_0_0'] = ''
            if 'TTYPE2' in hdulist[i].header and hdulist[i].header['TTYPE2'] == 'Delay':
                hdulist[i].header['TUNIT2'] = ''
                del hdulist[i].header['TUNIT2']
                hdulist[i].header['TUNIT2'] = 'ps'

            hdulist[i].verify('fix+warn')
            hdulist[i].header.update()
            # This actually requires substantially more work because it is lossy to information
            # on the unit that was encoded

        hdu = hdulist[1]

        scan_desc = copy.deepcopy(scan_desc)
        attrs = scan_desc.pop('note', scan_desc)
        attrs.update(dict(hdulist[0].header))

        drop_attrs = ['COMMENT', 'HISTORY', 'EXTEND', 'SIMPLE', 'SCANPAR', 'SFKE_0']
        for dropped_attr in drop_attrs:
            if dropped_attr in attrs:
                del attrs[dropped_attr]

        from arpes.utilities import rename_keys
        built_coords, dimensions, real_spectrum_shape = find_clean_coords(hdu, attrs, mode='MC')
        attrs = rename_keys(attrs, self.RENAME_KEYS)
        scan_desc = rename_keys(scan_desc, self.RENAME_KEYS)

        def clean_key_name(k: str) -> str:
            if '#' in k:
                k = k.replace('#', 'num')

            return k

        attrs = {clean_key_name(k): v for k, v in attrs.items()}
        scan_desc = {clean_key_name(k): v for k, v in scan_desc.items()}

        # don't have phi because we need to convert pixels first
        deg_to_rad_coords = {'beta', 'theta', 'chi'}

        # convert angular attributes to radians
        for coord_name in deg_to_rad_coords:
            if coord_name in attrs:
                try:
                    attrs[coord_name] = float(attrs[coord_name]) * (np.pi / 180)
                except (TypeError, ValueError):
                    pass
            if coord_name in scan_desc:
                try:
                    scan_desc[coord_name] = float(scan_desc[coord_name]) * (np.pi / 180)
                except (TypeError, ValueError):
                    pass

        data_vars = {}

        all_names = hdu.columns.names
        n_spectra = len([n for n in all_names if 'Fixed_Spectra' in n or 'Swept_Spectra' in n])
        for column_name in hdu.columns.names:
            # we skip some fixed set of the columns, such as the one dimensional axes, as well as things that are too
            # tricky to load at the moment, like the microscope images from MAESTRO
            should_skip = False
            if column_name in self.SKIP_COLUMN_NAMES:
                should_skip = True

            for formula in self.SKIP_COLUMN_FORMULAS:
                if formula(column_name):
                    should_skip = True

            if should_skip:
                continue

            # the hemisphere axis is handled below
            dimension_for_column = dimensions[column_name]
            column_shape = real_spectrum_shape[column_name]

            column_display = self.PREPPED_COLUMN_NAMES.get(column_name, column_name)
            if 'Fixed_Spectra' in column_display:
                if n_spectra == 1:
                    column_display = 'spectrum'
                else:
                    column_display = 'spectrum' + '-' + column_display.split('Fixed_Spectra')[1]

            if 'Swept_Spectra' in column_display:
                if n_spectra == 1:
                    column_display = 'spectrum'
                else:
                    column_display = 'spectrum' + '-' + column_display.split('Swept_Spectra')[1]

            # sometimes if a scan is terminated early it can happen that the sizes do not match the expected value
            # as an example, if a beta map is supposed to have 401 slices, it might end up having only 260 if it were
            # terminated early
            # If we are confident in our parsing code above, we can handle this case and take a subset of the coords
            # so that the data matches
            try:
                resized_data = hdu.data.columns[column_name].array.reshape(column_shape)
            except ValueError:
                # if we could not resize appropriately, we will try to reify the shapes together
                rest_column_shape = column_shape[1:]
                n_per_slice = int(np.prod(rest_column_shape))
                total_shape = hdu.data.columns[column_name].array.shape
                total_n = np.prod(total_shape)

                n_slices = total_n // n_per_slice
                # if this isn't true, we can't recover
                data_for_resize = hdu.data.columns[column_name].array
                if (total_n // n_per_slice != total_n / n_per_slice):
                    # the last slice was in the middle of writing when something hit the fan
                    # we need to infer how much of the data to read, and then repeat the above
                    # we need to cut the data

                    # This can happen when the labview crashes during data collection,
                    # we use column_shape[1] because of the row order that is used in the FITS file
                    data_for_resize = data_for_resize[0:(total_n // n_per_slice) * column_shape[1]]
                    warnings.warn(
                        'Column {} was in the middle of slice when DAQ stopped. Throwing out incomplete slice...'.format(
                            column_name))

                column_shape = list(column_shape)
                column_shape[0] = n_slices

                try:
                    resized_data = data_for_resize.reshape(column_shape)
                except Exception:
                    # sometimes for whatever reason FITS errors and cannot read the data
                    continue

                # we also need to adjust the coordinates
                altered_dimension = dimension_for_column[0]
                built_coords[altered_dimension] = built_coords[altered_dimension][:n_slices]

            data_vars[column_display] = xr.DataArray(
                resized_data,
                coords={k: c for k, c in built_coords.items() if k in dimension_for_column},
                dims=dimension_for_column,
                attrs=attrs,
            )

        def prep_spectrum(data: xr.DataArray):
            # don't do center pixel inference because the main chamber
            # at least consistently records the offset from the edge
            # of the recorded window
            if 'pixel' in data.coords:
                phi_axis = data.coords['pixel'].values * \
                           arpes.constants.SPECTROMETER_MC['rad_per_pixel']
                data = replace_coords(data, {
                    'phi': phi_axis
                }, [('pixel', 'phi',)])

            # Always attach provenance
            provenance_from_file(data, frame_path, {
                'what': 'Loaded MC dataset from FITS.',
                'by': 'load_MC',
            })

            return data

        if 'spectrum' in data_vars:
            data_vars['spectrum'] = prep_spectrum(data_vars['spectrum'])

        # adjust angular coordinates
        built_coords = {k: c * (np.pi / 180) if k in deg_to_rad_coords else c
                        for k, c in built_coords.items()}

        return xr.Dataset(
            data_vars,
            attrs={**scan_desc, 'name': primary_dataset_name},
        )


class SynchrotronEndstation(EndstationBase):
    """
    Synchrotron endstations have somewhat complicated light source metadata.
    This stub exists to attach commonalities, such as a resolution table which
    can be interpolated into to retrieve the x-ray linewidth at the
    experimental settings. Additionally, subclassing this is used in resolution
    calculations to signal that such a resolution lookup is required.
    """
    RESOLUTION_TABLE = None


class HemisphericalEndstation(EndstationBase):
    """
    An endstation definition for a hemispherical analyzer should include
    everything needed to determine energy + k resolution, angle conversion,
    and ideally correction databases for dead pixels + detector nonlinearity
    information
    """
    ANALYZER_INFORMATION = None
    SLIT_ORIENTATION = None
    PIXELS_PER_DEG = None


def endstation_from_alias(alias: str) -> type:
    """
    Lookup the data loading class from an alias.
    :param alias:
    :return:
    """
    return _ENDSTATION_ALIASES[alias]


def endstation_name_from_alias(alias):
    """
    Lookup the data loading principal location from an alias.
    :param alias:
    :return:
    """
    return endstation_from_alias(alias).PRINCIPAL_NAME


def add_endstation(endstation_cls: type) -> None:
    """
    Registers a data loading plugin (Endstation class) together with
    its aliases.

    :param endstation_cls:
    :return:
    """
    assert endstation_cls.PRINCIPAL_NAME is not None
    for alias in endstation_cls.ALIASES:
        if alias in _ENDSTATION_ALIASES:
            continue

        _ENDSTATION_ALIASES[alias] = endstation_cls

    if endstation_cls.PRINCIPAL_NAME in _ENDSTATION_ALIASES and endstation_cls.PRINCIPAL_NAME not in endstation_cls.ALIASES:
        # indicates it was added earlier, so there's an alias conflict
        if False:
            warnings.warn('Endstation name or alias conflicts with existing {}'.format(endstation_cls.PRINCIPAL_NAME))

    _ENDSTATION_ALIASES[endstation_cls.PRINCIPAL_NAME] = endstation_cls


def load_scan_for_endstation(scan_desc, endstation_cls, **kwargs):
    note = scan_desc.get('note', scan_desc)
    full_note = copy.deepcopy(scan_desc)
    full_note.update(note)

    return endstation_cls().load(scan_desc, **kwargs)


def load_scan(scan_desc: Dict[str, str], retry=True, **kwargs: Any) -> Dataset:
    """
    Determines which data loading class is appropriate for the data,
    shuffles a bit of metadata, and calls the .load function on the
    retrieved class to start the data loading process.
    :param scan_desc:
    :param retry: Used to attempt a reload of plugins and subsequent data load attempt.
    :param kwargs:
    :return:
    """
    note = scan_desc.get('note', scan_desc)
    full_note = copy.deepcopy(scan_desc)
    full_note.update(note)

    endstation_name = case_insensitive_get(full_note, 'location', case_insensitive_get(full_note, 'endstation'))
    try:
        endstation_cls = endstation_from_alias(endstation_name)
        return endstation_cls().load(scan_desc, **kwargs)
    except KeyError:
        if retry:
            import arpes.config  # pylint: disable=redefined-outer-name
            arpes.config.load_plugins()
            return load_scan(full_note, retry=False, **kwargs)
        else:
            raise ValueError('Could not identify endstation. '
                             'Did you set the endstation or location? Find a description of the available options '
                             'in the endstations module.')


import numpy as np

import xarray as xr
from arpes.analysis.path import select_along_path
from arpes.exceptions import AnalysisError
from arpes.plotting.interactive_utils import CursorTool, SaveableTool
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('PathTool', 'path_tool',)


class PathTool(SaveableTool, CursorTool):
    """
    Tool to allow drawing paths on data, creating selections based on paths,
    and masking regions around paths

    Integrates with the tools in arpes.analysis.path
    """

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop('name', None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get('app_main_size', 600)
        self.app_marginal_size = self.settings.get('app_main_size', 300)
        self.pointer_mode = 'path'

    def tool_handler(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column, widgetbox
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets, warnings
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError('Cannot use path tool on non image-like spectra')

        arr = self.arr
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        default_palette = self.default_palette

        self.app_context.update({
            'path_options': [],
            'active_path': None,
            'paths': {},
            'data': arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
            },
        })

        self.cursor = [np.mean(self.data_range['x']),
                       np.mean(self.data_range['y'])]

        self.color_maps['main'] = LinearColorMapper(default_palette, low=np.min(arr.values), high=np.max(arr.values),
                                                    nan_color='black')

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = 'Path Tool: WARNING Unidentified'

        try:
            main_title = 'Path Tool: {}'.format(arr.S.label[:60])
        except:
            pass

        self.figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=20, toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.data_range['x'], y_range=self.data_range['y'],
        )

        self.figures['main'].xaxis.axis_label = arr.dims[0]
        self.figures['main'].yaxis.axis_label = arr.dims[1]

        self.plots['main'] = self.figures['main'].image(
            [np.asarray(arr.values.T)], x=self.data_range['x'][0], y=self.data_range['y'][0],
            dw=self.data_range['x'][1] - self.data_range['x'][0],
            dh=self.data_range['y'][1] - self.data_range['y'][0],
            color_mapper=self.color_maps['main']
        )

        self.plots['paths'] = self.figures['main'].multi_line(xs=[], ys=[], line_color='white', line_width=2)

        self.add_cursor_lines(self.figures['main'])

        def add_point_to_path():
            if self.active_path in self.paths:
                self.paths[self.active_path]['points'].append(list(self.cursor))
                update_path_display()

            self.save_app()

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            if self.pointer_mode == 'path':
                add_point_to_path()

        POINTER_MODES = [
            ('Cursor', 'cursor',),
            ('Path', 'path',),
        ]

        def convert_to_xarray():
            """
            For each of the paths, we will create a dataset which has an index dimension,
            and datavariables for each of the coordinate dimensions
            :return:
            """
            def convert_single_path_to_xarray(points):
                vars = {d: np.array([p[i] for p in points]) for i, d in enumerate(self.arr.dims)}
                coords = {
                    'index': np.array(range(len(points))),
                }
                vars = {k: xr.DataArray(v, coords=coords, dims=['index']) for k, v in vars.items()}
                return xr.Dataset(data_vars=vars, coords=coords)

            return {path['name']: convert_single_path_to_xarray(path['points'])
                    for path in self.paths.values()}

        def select(data=None, **kwargs):
            if data is None:
                data = self.arr

            if len(self.paths) > 1:
                warnings.warn('Only using first path.')

            return select_along_path(path=list(convert_to_xarray().items())[0][1],
                                     data=data, **kwargs)

        self.app_context['to_xarray'] = convert_to_xarray
        self.app_context['select'] = select


        pointer_dropdown = widgets.Dropdown(label='Pointer Mode', button_type='primary', menu=POINTER_MODES)
        self.path_dropdown = widgets.Dropdown(label='Active Path', button_type='primary',
                                              menu=self.path_options)

        path_name_input = widgets.TextInput(placeholder='Path name...')
        add_path_button = widgets.Button(label='Add Path')

        clear_path_button = widgets.Button(label='Clear Path')
        remove_path_button = widgets.Button(label='Remove Path')

        main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range')

        def add_path(path_name):
            if path_name not in self.paths:
                self.path_options.append((path_name, path_name,))
                self.path_dropdown.menu = self.path_options
                self.paths[path_name] = {
                    'points': [],
                    'name': path_name,
                }

                if self.active_path is None:
                    self.active_path = path_name

                self.save_app()

        def on_change_active_path(attr, old_path_id, path_id):
            self.debug_text = path_id
            self.app_context['active_path'] = path_id
            self.active_path = path_id
            self.save_app()

        def on_change_pointer_mode(attr, old_pointer_mode, pointer_mode):
            self.app_context['pointer_mode'] = pointer_mode
            self.pointer_mode = pointer_mode
            self.save_app()

        def update_path_display():
            self.plots['paths'].data_source.data = {
                'xs': [[point[0] for point in p['points']] for p in self.paths.values()],
                'ys': [[point[1] for point in p['points']] for p in self.paths.values()],
            }
            self.save_app()

        self.update_path_display = update_path_display

        def on_clear_path():
            if self.active_path in self.paths:
                self.paths[self.active_path]['points'] = []
                update_path_display()

        def on_remove_path():
            if self.active_path in self.paths:
                del self.paths[self.active_path]
                new_path_options = [b for b in self.path_options if b[0] != self.active_path]
                self.path_dropdown.menu = new_path_options
                self.path_options = new_path_options
                self.active_path = None
                update_path_display()

        # Attach callbacks
        main_color_range_slider.on_change('value', self.update_colormap_for('main'))

        self.figures['main'].on_event(events.Tap, click_main_image)
        self.path_dropdown.on_change('value', on_change_active_path)
        pointer_dropdown.on_change('value', on_change_pointer_mode)
        add_path_button.on_click(lambda: add_path(path_name_input.value))
        clear_path_button.on_click(on_clear_path)
        remove_path_button.on_click(on_remove_path)

        layout = row(column(self.figures['main']),
                     column(
                         widgetbox(
                             pointer_dropdown,
                             self.path_dropdown,
                         ),
                         row(
                             path_name_input,
                             add_path_button,
                         ),
                         row(
                             clear_path_button,
                             remove_path_button,
                         ),
                         widgetbox(
                             self._cursor_info,
                             main_color_range_slider,
                         ),
                         self.debug_div,
                     ))

        doc.add_root(layout)
        doc.title = 'Path Tool'
        self.load_app()
        self.save_app()

    def serialize(self):
        return {
            'active_path': self.active_path,
            'path_options': self.path_options,
            'paths': self.paths,
            'cursor': self.cursor,
        }

    def deserialize(self, json_data):
        self.cursor = json_data.get('cursor', [0, 0])

        self.app_context['paths'] = json_data.get('paths', {}) or {}
        self.app_context['path_options'] = json_data.get('path_options', [])

        self.path_dropdown.menu = self.app_context['path_options']

        self.update_path_display()


def path_tool(data: DataType, **kwargs):
    data = normalize_to_spectrum(data)

    tool = PathTool(**kwargs)
    return tool.make_tool(data)



"""
Implements Igor <-> xarray interop, notably loading Igor waves and
packed experiment files.
"""

import re
import typing
import warnings

from pathlib import Path
from typing import Any

import numpy as np
import xarray as xr

__all__ = ('read_single_pxt', 'read_separated_pxt', 'read_experiment',
           'find_ses_files_associated',)

binary_header_bytes = 10

igor_wave_header_dtype = np.dtype([
    ('next', '>u8',),
    ('creation_date', '>u8'),
    ('mod_date', '>u8'),
    ('type', '>H'),
    ('d_lock', '>H'),
    ('wave_header_pad1', 'B', (62,),), # unused
    ('n_points', '>u8'),
    ('wave_header_version', '>u4'),
    ('spacer', '>u8'),
    ('wave_name', 'S', 32),
    ('wave_header_pad2', 'B', (8,)),
    ('dim_sizes', '>u4', (4,)),
    ('dim_scales', '>d', (4,)),
    ('dim_offsets', '>d', (4,)),
    ('data_units', 'b', (4,),),
    ('dim_units', ('S', 4), (4,)),
    ('final_spacer', 'b', 2 + 8 * 18),
])


def read_igor_binary_wave(raw_bytes):
    """
    Some weirdness here with the byte ordering and the target datatype. Additionally,
    newer Igor wave formats are very poorly documented. Data loading might not be perfectly
    correct as a result. For most purposes, you can load from .pxp files anyway.

    :param raw_bytes:
    :return:
    """
    header = np.fromstring(raw_bytes[:igor_wave_header_dtype.itemsize],
                           dtype=igor_wave_header_dtype).item()

    point_size = 8
    n_points = header[6]
    _ = header[9] # wave_name
    dim_sizes, dim_scales, dim_offsets = header[11], header[12], header[13]
    _ = header[14] # data_units
    dim_units = header[15]

    # approximate offsets
    # >i8 -1
    # <i8 0
    # >u8 -1
    # <u8 0
    offset = -1

    wave_data = np.fromstring(
        raw_bytes[igor_wave_header_dtype.itemsize + offset:igor_wave_header_dtype.itemsize + n_points * point_size + offset],
        dtype='>u8')

    wave_data = wave_data / np.max(wave_data)
    wave_data[wave_data == 1] = 0 # some weird bit shifting going on?

    names_from_units = {
        'eV': 'eV',
        'deg': 'phi',
    }

    dim_sizes = [d for d in dim_sizes if d]
    n_dims = len(dim_sizes)

    # please pylint forgive me
    dims = [(names_from_units.get(dim_units[i].decode('ascii'), dim_units[i].decode('ascii')),
             np.linspace(dim_offsets[i], dim_offsets[i] + (dim_sizes[i] - 1) * dim_scales[i],
                         dim_sizes[i])) for i in range(n_dims)]
    coords = dict(dims)

    return xr.DataArray(
        wave_data.reshape(dim_sizes[::-1]),
        coords=coords,
        dims=[d[0] for d in dims][::-1],
    )


def read_header(header_bytes: bytes):
    header_as_string = header_bytes.decode('utf-8')
    lines = [x for x in header_as_string.replace('\r', '\n').split('\n') if x]
    lines = [x for x in lines if '=' in x]

    header = {}
    for line in lines:
        fragments = line.split('=')
        first, rest = fragments[0], '='.join(fragments[1:])

        try:
            rest = int(rest)
        except ValueError:
            try:
                rest = float(rest)
            except ValueError:
                pass

        header[first.lower().replace(' ', '_')] = rest

    from arpes.utilities import rename_keys
    return rename_keys(header, {
        'sample_x': 'x',
        'sample_y_(vert)': 'y',
        'sample_y': 'y',
        'sample_z': 'z',
        'bl_energy': 'hv',
    })


def wave_to_xarray(wave: Any) -> xr.DataArray: # wave: igor.Wave
    """
    Converts a wave to an xarray.DataArray
    :param w:
    :return:
    """

    # only need four because Igor only supports four dimensions!
    extra_names = iter(['W', 'X', 'Y', 'Z'])
    n_dims = len([a for a in wave.axis if len(a)])

    def get_axis_name(index: int) -> str:
        unit = wave.axis_units[index]
        if unit:
            return {
                'eV': 'eV',
                'deg': 'phi',
                'Pwr Supply V':'volts',
                'K2200 V':'volts',
            }.get(unit, unit)

        return next(extra_names)

    axis_names = [get_axis_name(i) for i in range(n_dims)]
    coords = dict(zip(axis_names, wave.axis))

    return xr.DataArray(
        wave.data,
        coords=coords,
        dims=axis_names,
        attrs=read_header(wave.notes),
    )


def read_experiment(reference_path: typing.Union[Path, str], **kwargs):
    """
    Reads a whole experiment and translates all contained waves into xr.Dataset instances as appropriate

    :param reference_path:
    :return:
    """
    import igor.igorpy as igor

    if isinstance(reference_path, Path):
        reference_path = str(reference_path.absolute())

    return igor.load(reference_path, **kwargs)


def read_single_ibw(reference_path: typing.Union[Path, str]):
    """
    Currently igorpy does not support this though
    Uses igor.igorpy to load an .ibw file
    :param reference_path:
    :return:
    """
    import igor.igorpy as igor

    if isinstance(reference_path, Path):
        reference_path = str(reference_path.absolute())
    return igor.load(reference_path)


def read_single_pxt(reference_path: typing.Union[Path, str], byte_order=None):
    """
    Uses igor.igorpy to load a single .PXT or .PXP file
    :return:
    """
    import igor.igorpy as igor

    if isinstance(reference_path, Path):
        reference_path = str(reference_path.absolute())

    loaded = None
    if byte_order is None:
        for try_byte_order in ['>', '=', '<']:
            try:
                loaded = igor.load(reference_path, initial_byte_order=try_byte_order)
                break
            except Exception: # pylint: disable=broad-except
                # bad byte ordering probably
                pass
    else:
        loaded = igor.load(reference_path, initial_byte_order=byte_order)

    children = [c for c in loaded.children if isinstance(c, igor.Wave)]

    if len(children) > 1:
        warnings.warn('Igor PXT file contained {} waves. Ignoring all but first.', len(children))

    return wave_to_xarray(children[0])


def read_single_pxt_old(reference_path: Path, separator=None):
    bytes_for_file = reference_path.read_bytes()

    fallbacks = [
        b'[BCS]',
        b'[SES]',
    ]

    if separator is None:
        for fallback in fallbacks:
            if fallback in bytes_for_file:
                separator = fallback

        if separator is None:
            raise ValueError('Could not find appropriate separator for file.')

    all_sections = bytes_for_file.split(separator)
    content, header = all_sections[0], separator.join(all_sections[1:])
    header = read_header(header)

    return content, header
    #wave = read_igor_binary_wave(content[10:])
    #wave.attrs.update(header)
    #return wave


def find_ses_files_associated(reference_path: Path, separator: str = 'S'):
    """
    SES Software creates a series of PXT files they are all sequenced with _S[0-9][0-9][0-9].pxt
    `find_ses_files_associated` will collect all the files in the sequence
    pointed to by `reference_path`
    :param reference_path:
    :param separator:
    :return:
    """
    name_match = re.match(r'([\w+]+)[{}][0-9][0-9][0-9]\.pxt'.format(separator), reference_path.name)

    if name_match is None:
        return [reference_path]

    # otherwise need to collect all of the components
    fragment = name_match.groups()[0]
    components = list(reference_path.parent.glob('{}*.pxt'.format(fragment)))
    components.sort()

    return components


def read_separated_pxt(reference_path: Path, separator=None, byte_order=None):
    # determine if separated or not
    components = find_ses_files_associated(reference_path)
    frames = [read_single_pxt(f, byte_order=byte_order) for f in components]

    if len(frames) == 1:
        return frames[0]

    # adjust as needed
    scan_coords = ['hv', 'polar', 'timed_power', 'tilt', 'volts']

    scan_coord = None
    max_different_values = -np.inf
    for possible_scan_coord in scan_coords:
        coordinates = [f.attrs[possible_scan_coord] for f in frames]
        n_different_values = len(set(coordinates))
        if n_different_values > max_different_values:
            max_different_values = n_different_values
            scan_coord = possible_scan_coord

    for f in frames:
        f.coords[scan_coord] = f.attrs[scan_coord]

    frames.sort(key=lambda x: x.coords[scan_coord])
    return xr.concat(frames, scan_coord)



"""
This is another core part of PyARPES. It provides a lot of extensions to
what comes out of the box in xarray. Some of these are useful generics,
generally on the .T extension, others collect and manipulate metadata,
interface with plotting routines, provide functional programming utilities,
etc.

If `f` is an ARPES spectrum, then `f.S` should provide a nice representation of your data
in a Jupyter cell. This is a complement to the text based approach that merely printing `f`
offers.
"""

import collections
import copy
import itertools
import warnings
from collections import OrderedDict
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from scipy import ndimage as ndi

import arpes.constants
import arpes.plotting as plotting

from arpes.analysis import rebin
from arpes.analysis.band_analysis_utils import (param_getter,
                                                param_stderr_getter)
from arpes.io import load_dataset_attrs
from arpes.models.band import MultifitBand
from arpes.plotting import BandTool, CurvatureTool, FitCheckTool, ImageTool
from arpes.plotting.utils import fancy_labels, remove_colorbars
from arpes.typing import DataType
from arpes.utilities import apply_dataarray
from arpes.utilities.conversion import slice_along_path
from arpes.utilities.math import shift_by
from arpes.utilities.region import DesignatedRegions, normalize_region


__all__ = ['ARPESDataArrayAccessor', 'ARPESDatasetAccessor', 'ARPESFitToolsAccessor']


def _iter_groups(grouped):
    """
    Iterates through a flattened sequence
    :param grouped:
    :return:
    """
    for k, value_or_list in grouped.items():
        try:
            for list_item in value_or_list:
                yield k, list_item
        except TypeError:
            yield k, value_or_list


def unwrap_xarray_item(item):
    try:
        return item.item()
    except (AttributeError, ValueError):
        return item


def unwrap_xarray_dict(d):
    return {k: unwrap_xarray_item(v) for k, v in d.items()}


class ARPESAccessorBase:
    """
    Base class for the xarray extensions that we put onto our datasets to make working with ARPES data a
    little cleaner. This allows you to access common attributes
    """

    def along(self, directions, **kwargs):
        return slice_along_path(self._obj, directions, **kwargs)

    def find(self, name):
        return [n for n in dir(self) if name in n]

    @property
    def sherman_function(self):
        # TODO Need to improve spin-metadata

        # for option in ['sherman', 'sherman_function', 'SHERMAN']:
        #     if option in self._obj.attrs:
        #         return self._obj.attrs[option]

        return 0.2

    @property
    def experimental_conditions(self):
        try:
            temp = self.temp
        except AttributeError:
            temp = None

        return {
            'hv': self.hv,
            'polarization': self.polarization,
            'temp': temp,
        }

    @property
    def polarization(self):
        if 'epu_pol' in self._obj.attrs:
            # merlin: TODO normalize these
            # check and complete
            try:
                return {
                    0: 'p',
                    1: 'rc',
                    2: 's',
                }.get(int(self._obj.attrs['epu_pol']))
            except ValueError:
                return self._obj.attrs['epu_pol']

        return None

    @property
    def is_subtracted(self):
        if self._obj.attrs.get('subtracted'):
            return True

        if isinstance(self._obj, xr.DataArray):
            # if at least 5% of the values are < 0 we should consider the data
            # to be best represented by a coolwarm map
            return (((self._obj < 0) * 1).mean() > 0.05).item()

    @property
    def is_spatial(self):
        """
        Infers whether a given scan has real-space dimensions and corresponds to
        SPEM or u/nARPES.
        :return:
        """
        if self.spectrum_type in {'ucut', 'spem'}:
            return True

        return any(d in {'X', 'Y', 'Z'} for d in self._obj.dims)

    @property
    def is_kspace(self):
        """
        Infers whether the scan is k-space converted or not. Because of the way this is defined, it will return
        true for XPS spectra, which I suppose is true but trivially.
        :return:
        """
        return not any(d in {'phi', 'theta', 'beta', 'angle'} for d in self._obj.dims)

    @property
    def is_slit_vertical(self):
        return np.abs(self.lookup_offset_coord('alpha') - np.pi / 2) < (np.pi / 180)

    @property
    def endstation(self):
        return self._obj.attrs['location']

    @property
    def is_synchrotron(self):
        endstation = self.endstation

        synchrotron_endstations = {
            'BL403',
            'BL702',
        }

        return endstation in synchrotron_endstations

    @property
    def hv(self):
        if 'hv' in self._obj.attrs:
            value = float(self._obj.attrs['hv'])
            if not np.isnan(value):
                return value

        if 'location' in self._obj.attrs:
            if self._obj.attrs['location'] == 'ALG-MC':
                return 5.93

        return None

    def fetch_ref_attrs(self):
        if 'ref_attrs' in self._obj.attrs:
            return self._obj.attrs
        if 'ref_id' in self._obj.attrs:
            self._obj.attrs['ref_attrs'] = load_dataset_attrs(self._obj.attrs['ref_id'])

        try:
            df = self._obj.attrs['df']
            ref_id = df[df.id == self.original_id].iloc[0].ref_id
            return load_dataset_attrs(ref_id)
        except Exception:
            return {}

    @property
    def scan_type(self):
        return None

    @property
    def spectrum_type(self):
        if 'spectrum_type' in self._obj.attrs and self._obj.attrs['spectrum_type']:
            return self._obj.attrs['spectrum_type']

        dim_types = {
            ('eV',): 'xps_spectrum',
            ('eV', 'phi'): 'spectrum',

            # this should check whether the other angular axis perpendicular to scan axis?
            ('eV', 'phi', 'beta'): 'map',
            ('eV', 'phi', 'theta'): 'map',

            ('eV', 'hv', 'phi'): 'hv_map',

            # kspace
            ('eV', 'kp'): 'spectrum',
            ('eV', 'kx', 'ky'): 'map',
            ('eV', 'kp', 'kz'): 'hv_map',
        }

        dims = tuple(sorted(list(self._obj.dims)))

        return dim_types.get(dims)

    @property
    def is_differentiated(self):
        history = self.short_history()
        return 'dn_along_axis' in history or 'curvature' in history

    def transpose_to_front(self, dim):
        dims = list(self._obj.dims)
        assert dim in dims
        dims.remove(dim)
        return self._obj.transpose(*([dim] + dims))

    def transpose_to_back(self, dim):
        dims = list(self._obj.dims)
        assert dim in dims
        dims.remove(dim)
        return self._obj.transpose(*(dims + [dim]))

    def select_around_data(self, points, radius=None, fast=False, safe=True, mode='sum', **kwargs):
        """
        Can be used to perform a selection along one axis as a function of another, integrating a region
        in the other dimensions. As an example, suppose we have a dataset with dimensions ('eV', 'kp', 'T',)
        and we also by fitting determined the Fermi momentum as a function of T, kp_F('T'), stored in the
        dataarray kFs. Then we could select momentum integrated EDCs in a small window around the fermi momentum
        for each temperature by using

        ```python
        edcs_at_fermi_momentum = full_data.S.select_around_data({'kp': kFs}, radius={'kp': 0.04}, fast=True)
        ```

        The resulting data will be EDCs for each T, in a region of radius 0.04 inverse angstroms around the
        Fermi momentum.

        :param points:
        :param radius:
        :param fast:
        :param safe:
        :param kwargs:
        :return:
        """
        if isinstance(self._obj, xr.Dataset):
            raise TypeError('Cannot use select_around on Datasets only DataArrays!')

        if mode not in {'sum', 'mean'}:
            raise ValueError('mode parameter should be either sum or mean.')

        if isinstance(points, (tuple, list,)):
            warnings.warn('Dangerous iterable points argument to `select_around`')
            points = dict(zip(points, self._obj.dims))
        if isinstance(points, xr.Dataset):
            points = {k: points[k].item() for k in points.data_vars}

        default_radii = {
            'kp': 0.02,
            'kz': 0.05,
            'phi': 0.02,
            'beta': 0.02,
            'theta': 0.02,
            'eV': 0.05,
            'delay': 0.2,
            'T': 2,
            'temp': 2,
        }

        unspecified = 0.1

        if isinstance(radius, float):
            radius = {d: radius for d in points.keys()}
        else:
            collected_terms = set('{}_r'.format(k) for k in points.keys()).intersection(
                set(kwargs.keys()))
            if collected_terms:
                radius = {d: kwargs.get('{}_r'.format(d), default_radii.get(d, unspecified))
                          for d in points.keys()}
            elif radius is None:
                radius = {d: default_radii.get(d, unspecified) for d in points.keys()}

        assert isinstance(radius, dict)
        radius = {d: radius.get(d, default_radii.get(d, unspecified)) for d in points.keys()}

        along_dims = list(points.values())[0].dims
        selected_dims = list(points.keys())

        stride = self._obj.T.stride(generic_dim_names=False)

        new_dim_order = [d for d in self._obj.dims if d not in along_dims] + list(along_dims)

        data_for = self._obj.transpose(*new_dim_order)
        new_data = data_for.sum(selected_dims, keep_attrs=True)
        for coord, value in data_for.T.iterate_axis(along_dims):
            nearest_sel_params = {}
            if safe:
                for d, v in radius.items():
                    if v < stride[d]:
                        nearest_sel_params[d] = points[d].sel(**coord)

                radius = {d: v for d, v in radius.items() if d not in nearest_sel_params}

            if fast:
                selection_slices = {d: slice(points[d].sel(**coord) - radius[d], points[d].sel(**coord) + radius[d])
                                    for d in points.keys() if d in radius}
                selected = value.sel(**selection_slices)
            else:
                raise NotImplementedError()

            if nearest_sel_params:
                selected = selected.sel(**nearest_sel_params, method='nearest')

            for d in nearest_sel_params:
                # need to remove the extra dims from coords
                del selected.coords[d]

            if mode == 'sum':
                new_data.loc[coord] = selected.sum(list(radius.keys())).values
            elif mode == 'mean':
                new_data.loc[coord] = selected.mean(list(radius.keys())).values

        return new_data


    def select_around(self, point, radius=None, fast=False, safe=True, mode='sum', **kwargs):
        """
        Selects and integrates a region around a one dimensional point, useful to do a small
        region integration, especially around points on a path of a k-point of interest.

        If the fast flag is set, we will use the Manhattan norm, i.e. sum over square regions
        rather than ellipsoids, as this is less costly.

        If radii are not set, or provided through kwargs as 'eV_r' or 'phi_r' for instance,
        then we will try to use reasonable default values; buyer beware.
        :param point:
        :param radius:
        :param fast:
        :return:
        """
        if isinstance(self._obj, xr.Dataset):
            raise TypeError('Cannot use select_around on Datasets only DataArrays!')

        if mode not in {'sum', 'mean'}:
            raise ValueError('mode parameter should be either sum or mean.')

        if isinstance(point, (tuple, list,)):
            warnings.warn('Dangerous iterable point argument to `select_around`')
            point = dict(zip(point, self._obj.dims))
        if isinstance(point, xr.Dataset):
            point = {k: point[k].item() for k in point.data_vars}

        default_radii = {
            'kp': 0.02,
            'kz': 0.05,
            'phi': 0.02,
            'beta': 0.02,
            'theta': 0.02,
            'eV': 0.05,
            'delay': 0.2,
            'T': 2,
        }

        unspecified = 0.1

        if isinstance(radius, float):
            radius = {d: radius for d in point.keys()}
        else:
            collected_terms = set('{}_r'.format(k) for k in point.keys()).intersection(
                set(kwargs.keys()))
            if collected_terms:
                radius = {d: kwargs.get('{}_r'.format(d), default_radii.get(d, unspecified))
                          for d in point.keys()}
            elif radius is None:
                radius = {d: default_radii.get(d, unspecified) for d in point.keys()}

        assert isinstance(radius, dict)
        radius = {d: radius.get(d, default_radii.get(d, unspecified)) for d in point.keys()}

        # make sure we are taking at least one pixel along each
        nearest_sel_params = {}
        if safe:
            stride = self._obj.T.stride(generic_dim_names=False)
            for d, v in radius.items():
                if v < stride[d]:
                    nearest_sel_params[d] = point[d]

            radius = {d: v for d, v in radius.items() if d not in nearest_sel_params}

        if fast:
            selection_slices = {d: slice(point[d] - radius[d], point[d] + radius[d])
                                for d in point.keys() if d in radius}
            selected = self._obj.sel(**selection_slices)
        else:
            # selected = self._obj
            raise NotImplementedError()

        if nearest_sel_params:
            selected = selected.sel(**nearest_sel_params, method='nearest')

        for d in nearest_sel_params:
            # need to remove the extra dims from coords
            del selected.coords[d]

        if mode == 'sum':
            return selected.sum(list(radius.keys()))
        elif mode == 'mean':
            return selected.mean(list(radius.keys()))

    def short_history(self, key='by'):
        return [h['record'][key] if isinstance(h, dict) else h for h in self.history]

    def _calculate_symmetry_points(self, symmetry_points, projection_distance=0.05,
                                   include_projected=True, epsilon=0.01):
        # For each symmetry point, we need to determine if it is projected or not
        # if it is projected, we need to calculate its projected coordinates
        points = collections.defaultdict(list)
        projected_points = collections.defaultdict(list)

        fixed_coords = {k: v for k, v in self._obj.coords.items() if k not in self._obj.indexes}
        index_coords = self._obj.indexes

        for point, locations in symmetry_points.items():
            if not isinstance(locations, list):
                locations = [locations]

            for location in locations:
                # determine whether the location needs to be projected
                projected = False
                skip = False
                for axis, value in location.items():
                    if axis in fixed_coords and np.abs(value - fixed_coords[axis]) > epsilon:
                        projected = True
                    if axis not in fixed_coords and axis not in index_coords:
                        # cannot even hope to do anything here, we don't have enough info
                        skip = True

                if skip:
                    continue

                new_location = location.copy()
                if projected:
                    # Go and do the projection, for now we will assume we just get it by
                    # replacing the value of the mismatched coordinates.

                    # This does not work if the coordinate system is not orthogonal
                    for axis, v in location.items():
                        if axis in fixed_coords:
                            new_location = fixed_coords[axis]

                    projected_points[point].append(location)
                else:
                    points[point].append(location)

        return points, projected_points

    def symmetry_points(self, raw=False, **kwargs):
        symmetry_points = self.fetch_ref_attrs().get('symmetry_points', {})
        our_symmetry_points = self._obj.attrs.get('symmetry_points', {})

        symmetry_points = copy.deepcopy(symmetry_points)
        symmetry_points.update(our_symmetry_points)

        if raw:
            return symmetry_points

        return self._calculate_symmetry_points(symmetry_points, **kwargs)

    @property
    def iter_own_symmetry_points(self):
        sym_points, _ = self.symmetry_points()
        return _iter_groups(sym_points)

    @property
    def iter_projected_symmetry_points(self):
        _, sym_points = self.symmetry_points()
        return _iter_groups(sym_points)

    @property
    def iter_symmetry_points(self):
        for sym_point in self.iter_own_symmetry_points:
            yield sym_point
        for sym_point in self.iter_projected_symmetry_points:
            yield sym_point

    @property
    def history(self):
        provenance = self._obj.attrs.get('provenance', None)

        def unlayer(prov):
            if prov is None:
                return [], None
            if isinstance(prov, str):
                return [prov], None
            first_layer = copy.copy(prov)

            rest = first_layer.pop('parents_provenance', None)
            if rest is None:
                rest = first_layer.pop('parents_provanence', None)
            if isinstance(rest, list):
                warnings.warn('Encountered multiple parents in history extraction, '
                              'throwing away all but the first.')
                if rest:
                    rest = rest[0]
                else:
                    rest = None

            return [first_layer], rest

        def _unwrap_provenance(prov):
            if prov is None:
                return []

            first, rest = unlayer(prov)

            return first + _unwrap_provenance(rest)

        return _unwrap_provenance(provenance)

    @property
    def spectrometer(self):
        ds = self._obj
        spectrometers = {
            'SToF': arpes.constants.SPECTROMETER_SPIN_TOF,
            'ToF': arpes.constants.SPECTROMETER_STRAIGHT_TOF,
            'DLD': arpes.constants.SPECTROMETER_DLD,
            'BL7': arpes.constants.SPECTROMETER_BL7,
            'ANTARES': arpes.constants.SPECTROMETER_ANTARES,
        }

        if 'spectrometer_name' in ds.attrs:
            return spectrometers.get(ds.attrs['spectrometer_name'])

        if isinstance(ds, xr.Dataset):
            if 'up' in ds.data_vars or ds.attrs.get("18  MCP3") == 0:
                return spectrometers['SToF']
        elif isinstance(ds, xr.DataArray):
            if ds.name == 'up' or ds.attrs.get("18  MCP3") == 0:
                return spectrometers['SToF']

        if 'location' in ds.attrs:
            return {
                'ALG-MC': arpes.constants.SPECTROMETER_MC,
                'BL403': arpes.constants.SPECTROMETER_BL4,
                'ALG-SToF': arpes.constants.SPECTROMETER_STRAIGHT_TOF,
                'Kaindl': arpes.constants.SPECTROMETER_KAINDL,
                'BL7': arpes.constants.SPECTROMETER_BL7,
                'ANTARES': arpes.constants.SPECTROMETER_ANTARES,
            }.get(ds.attrs['location'])

        try:
            return spectrometers[ds.attrs['spectrometer_name']]
        except KeyError:
            return {}

    @property
    def dshape(self):
        arr = self._obj
        return dict(zip(arr.dims, arr.shape))

    @property
    def original_id(self):
        history = self.history
        if len(history) >= 3:
            first_modification = history[-3]
            return first_modification['parent_id']

        return self._obj.attrs['id']

    @property
    def original_parent_scan_name(self):
        try:
            history = self.history
            if len(history) >= 3:
                first_modification = history[-3]
                df = self._obj.attrs['df']
                return df[df.id == first_modification['parent_id']].index[0]
        except:
            pass
        return ""

    @property
    def scan_row(self):
        df = self._obj.attrs['df']
        sdf = df[df.path == self._obj.attrs['file']]
        return list(sdf.iterrows())[0]

    @property
    def df_index(self):
        return self.scan_row[0]

    @property
    def df_after(self):
        return self._obj.attrs['df'][self._obj.attrs['df'].index > self.df_index]

    def df_until_type(self, df=None, spectrum_type=None):
        if df is None:
            df = self.df_after

        if spectrum_type is None:
            spectrum_type = (self.spectrum_type,)

        if isinstance(spectrum_type, str):
            spectrum_type = (spectrum_type,)

        try:
            indices = [df[df['spectrum_type'].eq(s)] for s in spectrum_type]
            indices = [d.index[0] for d in indices if not d.empty]

            if not indices:
                raise IndexError()

            min_index = min(indices)
            return df[df.index < min_index]
        except IndexError:
            # nothing
            return df

    @property
    def scan_name(self):
        for option in ['scan', 'file']:
            if option in self._obj.attrs:
                return self._obj.attrs[option]

        id = self._obj.attrs.get('id')

        if id is None:
            return 'No ID'

        try:
            df = self._obj.attrs['df']
            return df[df.id == id].index[0]
        except (IndexError, KeyError, AttributeError):
            # data is probably not raw data
            return self.original_parent_scan_name

    @property
    def label(self):
        return str(self._obj.attrs.get('description', self.scan_name))

    @property
    def t0(self):
        if 't0' in self._obj.attrs:
            value = float(self._obj.attrs['t0'])
            if not np.isnan(value):
                return value

        if 'T0_ps' in self._obj.attrs:
            value = float(self._obj.attrs['T0_ps'])
            if not np.isnan(value):
                return value

        return None

    def apply_offsets(self, offsets):
        for k, v in offsets.items():
            self._obj.attrs['{}_offset'.format(k)] = v

    def lookup_offset_coord(self, name):
        offset = self.lookup_coord(name) - self.lookup_offset(name)
        try:
            return offset.item()
        except (AttributeError, ValueError):
            try:
                return offset.values
            except AttributeError:
                return offset

    def lookup_coord(self, name):
        if name in self._obj.coords:
            return self._obj.coords[name]

        if name in self._obj.attrs:
            return self._obj.attrs[name]

        raise ValueError('Could not find coordinate {}.'.format(name))

    def lookup_offset(self, attr_name):
        symmetry_points = self.symmetry_points(raw=True)
        if 'G' in symmetry_points:
            gamma_point = symmetry_points['G']
            if attr_name in gamma_point:
                return gamma_point[attr_name]

        offset_name = attr_name + '_offset'
        if offset_name in self._obj.attrs:
            return self._obj.attrs[offset_name]

        return self._obj.attrs.get('data_preparation', {}).get(offset_name, 0)

    @property
    def beta_offset(self):
        return self.lookup_offset('beta')

    @property
    def theta_offset(self):
        return self.lookup_offset('theta')

    @property
    def phi_offset(self):
        return self.lookup_offset('phi')

    @property
    def work_function(self):
        """
        We should try to improve this.
        :return:
        """

        if 'sample_workfunction' in self._obj.attrs:
            return self._obj.attrs['sample_workfunction']

        return 4.32

    @property
    def inner_potential(self):
        """
        We should improve this as well.
        :return:
        """
        if 'inner_potential' in self._obj.attrs:
            return self._obj.attrs['inner_potential']

        return 10

    def find_spectrum_energy_edges(self, indices=False):
        energy_marginal = self._obj.sum([d for d in self._obj.dims if d not in ['eV']])

        embed_size = 20
        embedded = np.ndarray(shape=[embed_size] + list(energy_marginal.values.shape))
        embedded[:] = energy_marginal.values
        embedded = ndi.gaussian_filter(embedded, embed_size / 3)

        # try to avoid dependency conflict with numpy v0.16
        from skimage import feature # pylint: disable=import-error
        edges = feature.canny(embedded, sigma=embed_size / 5, use_quantiles=True,
                              low_threshold=0.3) * 1
        edges = np.where(edges[int(embed_size / 2)] == 1)[0]
        if indices:
            return edges

        delta = self._obj.T.stride(generic_dim_names=False)
        return edges * delta['eV'] + self._obj.coords['eV'].values[0]

    def find_spectrum_angular_edges_full(self, indices=False):
        # as a first pass, we need to find the bottom of the spectrum, we will use this
        # to select the active region and then to rebin into course steps in energy from 0
        # down to this region
        # we will then find the appropriate edge for each slice, and do a fit to the edge locations

        energy_edge = self.find_spectrum_energy_edges()
        low_edge = np.min(energy_edge) + 0.05
        high_edge = np.max(energy_edge) - 0.05

        if high_edge - low_edge < 0.15:
            # Doesn't look like the automatic inference of the energy edge was valid
            high_edge = 0
            low_edge = np.min(self._obj.coords['eV'].values)

        angular_dim = 'pixel' if 'pixel' in self._obj.dims else 'phi'
        energy_cut = self._obj.sel(eV=slice(low_edge, high_edge)).S.sum_other(['eV', angular_dim])

        n_cuts = int(np.ceil(high_edge - low_edge / 0.05))
        new_shape = {'eV': n_cuts}
        new_shape[angular_dim] = len(energy_cut.coords[angular_dim].values)
        rebinned = rebin(energy_cut, shape=new_shape)

        embed_size = 20
        embedded = np.ndarray(shape=[embed_size] + [len(rebinned.coords[angular_dim].values)])
        low_edges = []
        high_edges = []
        for e_cut in rebinned.coords['eV'].values:
            e_slice = rebinned.sel(eV=e_cut)
            values = e_slice.values
            values[values > np.mean(values)] = np.mean(values)
            embedded[:] = values
            embedded = ndi.gaussian_filter(embedded, embed_size / 1.5)

            # try to avoid dependency conflict with numpy v0.16
            from skimage import feature # pylint: disable=import-error
            edges = feature.canny(embedded, sigma=4, use_quantiles=False,
                                  low_threshold=0.7, high_threshold=1.5) * 1
            edges = np.where(edges[int(embed_size / 2)] == 1)[0]
            low_edges.append(np.min(edges))
            high_edges.append(np.max(edges))

        if indices:
            return np.array(low_edges), np.array(high_edges), rebinned.coords['eV']

        delta = self._obj.T.stride(generic_dim_names=False)

        low_edges = np.array(low_edges) * delta[angular_dim] + rebinned.coords[angular_dim].values[0]
        high_edges = np.array(high_edges) * delta[angular_dim] + rebinned.coords[angular_dim].values[0]

        return low_edges, high_edges, rebinned.coords['eV']

    def zero_spectrometer_edges(self, edge_type='hard', cut_margin=None, interp_range=None, low=None, high=None):
        """
        At the moment we only provide hard edges. Soft edges would be smoothed
        with a logistic function or similar.
        :param edge_type:
        :param inter_range: Range over which to extrapolate fit
        :return:
        """

        if low is not None:
            assert high is not None
            assert len(low) == len(high) == 2

            low_edges = low
            high_edges = high

        low_edges, high_edges, rebinned_eV_coord = self.find_spectrum_angular_edges_full(indices=True)




        angular_dim = 'pixel' if 'pixel' in self._obj.dims else 'phi'
        if cut_margin is None:
            if 'pixel' in self._obj.dims:
                cut_margin = 50
            else:
                cut_margin = int(0.08 / self._obj.T.stride(generic_dim_names=False)[angular_dim])
        else:
            if isinstance(cut_margin, float):
                assert angular_dim == 'phi'
                cut_margin = int(cut_margin / self._obj.T.stride(generic_dim_names=False)[angular_dim])

        if interp_range is not None:
            low_edge = xr.DataArray(low_edges, coords={'eV': rebinned_eV_coord}, dims=['eV'])
            high_edge = xr.DataArray(high_edges, coords={'eV': rebinned_eV_coord}, dims=['eV'])
            low_edge = low_edge.sel(eV=interp_range)
            high_edge = high_edge.sel(eV=interp_range)
            import pdb
            pdb.set_trace()

        other_dims = list(self._obj.dims)
        other_dims.remove('eV')
        other_dims.remove(angular_dim)
        copied = self._obj.copy(deep=True).transpose(*(['eV', angular_dim] + other_dims))

        low_edges += cut_margin
        high_edges -= cut_margin

        for i, energy in enumerate(copied.coords['eV'].values):
            index = np.searchsorted(rebinned_eV_coord, energy)
            other = index + 1
            if other >= len(rebinned_eV_coord):
                other = len(rebinned_eV_coord) - 1
                index = len(rebinned_eV_coord) - 2

            low = int(np.interp(energy, rebinned_eV_coord, low_edges))
            high = int(np.interp(energy, rebinned_eV_coord, high_edges))
            copied.values[i, 0:low] = 0
            copied.values[i, high:-1] = 0

        return copied

    def sum_other(self, dim_or_dims):
        if isinstance(dim_or_dims, str):
            dim_or_dims = [dim_or_dims]

        return self._obj.sum([d for d in self._obj.dims if d not in dim_or_dims])

    def find_spectrum_angular_edges(self, indices=False):
        angular_dim = 'pixel' if 'pixel' in self._obj.dims else 'phi'
        energy_edge = self.find_spectrum_energy_edges()
        energy_slice = slice(np.max(energy_edge) - 0.1, np.max(energy_edge))
        near_ef = self._obj.sel(eV=energy_slice).sum([d for d in self._obj.dims if d not in [angular_dim]])

        embed_size = 20
        embedded = np.ndarray(shape=[embed_size] + list(near_ef.values.shape))
        embedded[:] = near_ef.values
        embedded = ndi.gaussian_filter(embedded, embed_size / 3)

        # try to avoid dependency conflict with numpy v0.16
        from skimage import feature # pylint: disable=import-error
        edges = feature.canny(embedded, sigma=embed_size / 5, use_quantiles=True,
                              low_threshold=0.2) * 1
        edges = np.where(edges[int(embed_size / 2)] == 1)[0]
        if indices:
            return edges

        delta = self._obj.T.stride(generic_dim_names=False)
        return edges * delta[angular_dim] + self._obj.coords[angular_dim].values[0]

    def trimmed_selector(self):
        raise NotImplementedError()

    def wide_angle_selector(self, include_margin=True):
        edges = self.find_spectrum_angular_edges()
        low_edge, high_edge = np.min(edges), np.max(edges)

        # go and build in a small margin
        if include_margin:
            if 'pixels' in self._obj.dims:
                low_edge += 50
                high_edge -= 50
            else:
                low_edge += 0.05
                high_edge -= 0.05

        return slice(low_edge, high_edge)

    def narrow_angle_selector(self):
        raise NotImplementedError()

    def meso_effective_selector(self):
        energy_edge = self.find_spectrum_energy_edges()
        return slice(np.max(energy_edge) - 0.3, np.max(energy_edge) - 0.1)

    def region_sel(self, *regions):
        def process_region_selector(selector: Union[slice, DesignatedRegions], dimension_name: str):
            if isinstance(selector, slice):
                return selector

            # need to read out the region
            options = {
                'eV': (DesignatedRegions.ABOVE_EF,
                       DesignatedRegions.BELOW_EF,
                       DesignatedRegions.EF_NARROW,
                       DesignatedRegions.MESO_EF,

                       DesignatedRegions.MESO_EFFECTIVE_EF,
                       DesignatedRegions.ABOVE_EFFECTIVE_EF,
                       DesignatedRegions.BELOW_EFFECTIVE_EF,
                       DesignatedRegions.EFFECTIVE_EF_NARROW),

                'phi': (DesignatedRegions.NARROW_ANGLE,
                        DesignatedRegions.WIDE_ANGLE,
                        DesignatedRegions.TRIM_EMPTY),
            }

            options_for_dim = options.get(dimension_name, [d for d in DesignatedRegions])
            assert selector in options_for_dim

            # now we need to resolve out the region
            resolution_methods = {
                DesignatedRegions.ABOVE_EF: slice(0, None),
                DesignatedRegions.BELOW_EF: slice(None, 0),
                DesignatedRegions.EF_NARROW: slice(-0.1, 0.1),
                DesignatedRegions.MESO_EF: slice(-0.3, -0.1),
                DesignatedRegions.MESO_EFFECTIVE_EF: self.meso_effective_selector,

                # Implement me
                #DesignatedRegions.TRIM_EMPTY: ,
                DesignatedRegions.WIDE_ANGLE: self.wide_angle_selector,
                #DesignatedRegions.NARROW_ANGLE: self.narrow_angle_selector,
            }
            resolution_method = resolution_methods[selector]
            if isinstance(resolution_method, slice):
                return resolution_method
            if callable(resolution_method):
                return resolution_method()

            raise NotImplementedError('Unable to determine resolution method.')

        obj = self._obj

        def unpack_dim(dim_name):
            if dim_name == 'angular':
                return 'pixel' if 'pixel' in obj.dims else 'phi'

            return dim_name

        for region in regions:
            region = {unpack_dim(k): v for k, v in normalize_region(region).items()}

            # remove missing dimensions from selection for permissiveness
            # and to transparent composing of regions
            region = {k: process_region_selector(v, k)
                      for k, v in region.items() if k in obj.dims}
            obj = obj.sel(**region)

        return obj

    def fat_sel(self, widths=None, **kwargs):
        """
        Allows integrating a selection over a given width. The produced dataset
        will be normalized by dividing by the number of single slices integrated over.

        This can be used to produce temporary datasets that have reduced
        uncorrelated noise.
        :param widths: Override the widths for the slices. Resonable defaults are used otherwise
        :param kwargs: slice dict. Has the same function as xarray.DataArray.sel
        :return:
        """
        if widths is None:
            widths = {}

        default_widths = {
            'eV': 0.05,
            'phi': 2,
            'beta': 2,
            'theta': 2,
            'kx': 0.02,
            'ky': 0.02,
            'kp': 0.02,
            'kz': 0.1,
        }

        extra_kwargs = {k: v for k, v in kwargs.items() if k not in self._obj.dims}
        slice_kwargs = {k: v for k, v in kwargs.items() if k not in extra_kwargs}
        slice_widths = {k: widths.get(k, extra_kwargs.get(k + '_width', default_widths.get(k)))
                        for k in slice_kwargs}

        slices = {k: slice(v - slice_widths[k] / 2, v + slice_widths[k] / 2)
                  for k, v in slice_kwargs.items()}

        sliced = self._obj.sel(**slices)
        thickness = np.product([len(sliced.coords[k]) for k in slice_kwargs.keys()])
        normalized = sliced.sum(slices.keys(), keep_attrs=True) / thickness
        for k, v in slices.items():
            normalized.coords[k] = (v.start + v.stop) / 2
        normalized.attrs.update(self._obj.attrs.copy())
        return normalized

    @property
    def reference_settings(self):
        settings = self.spectrometer_settings or {}

        settings.update({
            'hv': self.hv,
        })

        return settings

    @property
    def beamline_settings(self):
        find_keys = {
            'entrance_slit': {'entrance_slit',},
            'exit_slit': {'exit_slit',},
            'hv': {'hv', 'photon_energy',},
            'grating': {},
        }
        settings = {}
        for key, options in find_keys.items():
            for option in options:
                if option in self._obj.attrs:
                    settings[key] = self._obj.attrs[option]
                    break

        if self.endstation == 'BL403':
            settings['grating'] = 'HEG' # for now assume we always use the first order light

        return settings

    @property
    def spectrometer_settings(self):
        find_keys = {
            'lens_mode': {'lens_mode',},
            'pass_energy': {'pass_energy', },
            'scan_mode': {'scan_mode',},
            'scan_region': {'scan_region',},
            'slit': {'slit', 'slit_plate',},
        }
        settings = {}
        for key, options in find_keys.items():
            for option in options:
                if option in self._obj.attrs:
                    settings[key] = self._obj.attrs[option]
                    break

        if isinstance(settings.get('slit'), (float, np.float32, np.float64)):
            settings['slit'] = int(round(settings['slit']))

        return settings

    @property
    def sample_pos(self):
        x, y, z = None, None, None
        try:
            x = self._obj.attrs['x']
        except KeyError:
            pass
        try:
            y = self._obj.attrs['y']
        except KeyError:
            pass
        try:
            z = self._obj.attrs['z']
        except KeyError:
            pass

        def do_float(w):
            return float(w) if w is not None else None

        return (do_float(x), do_float(y), do_float(z))

    @property
    def chi(self):
        """
        This angle is always the puck rotation angle, i.e. the angle that spins the puck about its center while keeping
        the manipulator body fixed. Other code sometimes refers to this angle as ``phi`` or ``sample_phi``
        :return:
        """
        options = ['chi', 'Azimuth']
        for option in options:
            if option in self._obj.attrs:
                return float(self._obj.attrs[option])

        return None

    @property
    def sample_angles(self):
        return (
            # manipulator
            self.lookup_coord('beta'),
            self.lookup_coord('theta'),
            self.lookup_coord('chi'),

            # analyzer
            self.lookup_coord('phi'),
            self.lookup_coord('psi'),
            self.lookup_coord('alpha'),
        )

    @property
    def full_coords(self):
        full_coords = {}

        full_coords.update(dict(zip(['x', 'y', 'z'], self.sample_pos)))
        full_coords.update(dict(zip(['beta', 'theta', 'chi', 'phi', 'psi', 'alpha'], self.sample_angles)))
        full_coords.update({
            'hv': self.hv,
        })

        full_coords.update(self._obj.coords)
        return full_coords

    @property
    def sample_info(self):
        return unwrap_xarray_dict({
            'id': self._obj.attrs.get('sample_id'),
            'name': self._obj.attrs.get('sample_name'),
            'source': self._obj.attrs.get('sample_source'),
            'reflectivity': self._obj.attrs.get('sample_reflectivity'),
        })

    @property
    def scan_info(self):
        return unwrap_xarray_dict({
            'time': self._obj.attrs.get('time'),
            'date': self._obj.attrs.get('date'),
            'type': self.scan_type,
            'spectrum_type': self.spectrum_type,
            'experimenter': self._obj.attrs.get('experimenter'),
            'sample': self._obj.attrs.get('sample_name'),
        })

    @property
    def experiment_info(self):
        return unwrap_xarray_dict({
            'temperature': self._obj.attrs.get('temperature'),
            'temperature_cryotip': self._obj.attrs.get('temperature_cryotip'),
            'pressure': self._obj.attrs.get('pressure'),
            'polarization': self.probe_polarization,
            'photon_flux': self._obj.attrs.get('photon_flux'),
            'photocurrent': self._obj.attrs.get('photocurrent'),
            'probe': self._obj.attrs.get('probe'),
            'probe_detail': self._obj.attrs.get('probe_detail'),
            'analyzer': self._obj.attrs.get('analyzer'),
            'analyzer_detail': self.analyzer_detail,
        })

    @property
    def pump_info(self):
        return unwrap_xarray_dict({
            'pump_wavelength': self._obj.attrs.get('pump_wavelength'),
            'pump_energy': self._obj.attrs.get('pump_energy'),
            'pump_fluence': self._obj.attrs.get('pump_fluence'),
            'pump_pulse_energy': self._obj.attrs.get('pump_pulse_energy'),
            'pump_spot_size': (self._obj.attrs.get('pump_spot_size_x'),
                               self._obj.attrs.get('pump_spot_size_y')),
            'pump_profile': self._obj.attrs.get('pump_profile'),
            'pump_linewidth': self._obj.attrs.get('pump_linewidth'),
            'pump_temporal_width': self._obj.attrs.get('pump_temporal_width'),
            'pump_polarization': self.pump_polarization,
        })

    @property
    def probe_info(self):
        return unwrap_xarray_dict({
            'probe_wavelength': self._obj.attrs.get('probe_wavelength'),
            'probe_energy': self._obj.coords['hv'],
            'probe_fluence': self._obj.attrs.get('probe_fluence'),
            'probe_pulse_energy': self._obj.attrs.get('probe_pulse_energy'),
            'probe_spot_size': (self._obj.attrs.get('probe_spot_size_x'),
                                self._obj.attrs.get('probe_spot_size_y')),
            'probe_profile': self._obj.attrs.get('probe_profile'),
            'probe_linewidth': self._obj.attrs.get('probe_linewidth'),
            'probe_temporal_width': self._obj.attrs.get('probe_temporal_width'),
            'probe_polarization': self.probe_polarization,
        })

    @property
    def laser_info(self):
        return {
            **self.probe_info,
            **self.pump_info,
            'repetition_rate': self._obj.attrs.get('repetition_rate'),
        }

    @property
    def analyzer_info(self):
        return unwrap_xarray_dict({
            'lens_mode': self._obj.attrs.get('lens_mode'),
            'lens_mode_name': self._obj.attrs.get('lens_mode_name'),
            'acquisition_mode': self._obj.attrs.get('acquisition_mode'),
            'pass_energy': self._obj.attrs.get('pass_energy'),
            'slit_shape': self._obj.attrs.get('slit_shape'),
            'slit_width': self._obj.attrs.get('slit_width'),
            'slit_number': self._obj.attrs.get('slit_number'),
            'lens_table': self._obj.attrs.get('lens_table'),
            'analyzer_type': self._obj.attrs.get('analyzer_type'),
            'mcp_voltage': self._obj.attrs.get('mcp_voltage'),
        })

    @property
    def daq_info(self):
        return unwrap_xarray_dict({
            'daq_type': self._obj.attrs.get('daq_type'),
            'region': self._obj.attrs.get('daq_region'),
            'region_name': self._obj.attrs.get('daq_region_name'),
            'center_energy': self._obj.attrs.get('daq_center_energy'),
            'prebinning': self.prebinning,
            'trapezoidal_correction_strategy': self._obj.attrs.get('trapezoidal_correction_strategy'),
            'dither_settings': self._obj.attrs.get('dither_settings'),
            'sweep_settings': self.sweep_settings,
            'frames_per_slice': self._obj.attrs.get('frames_per_slice'),
            'frame_duration': self._obj.attrs.get('frame_duration'),
        })

    @property
    def beamline_info(self):
        return unwrap_xarray_dict({
            'hv': self._obj.coords['hv'],
            'linewidth': self._obj.attrs.get('probe_linewidth'),
            'photon_polarization': self.probe_polarization,
            'undulator_info': self.undulator_info,
            'repetition_rate': self._obj.attrs.get('repetition_rate'),
            'beam_current': self._obj.attrs.get('beam_current'),
            'entrance_slit': self._obj.attrs.get('entrance_slit'),
            'exit_slit': self._obj.attrs.get('exit_slit'),
            'monochromator_info': self.monochromator_info,
        })

    @property
    def sweep_settings(self):
        return {
            'high_energy': self._obj.attrs.get('sweep_high_energy'),
            'low_energy': self._obj.attrs.get('sweep_low_energy'),
            'n_sweeps': self._obj.attrs.get('n_sweeps'),
            'step': self._obj.attrs.get('sweep_step'),
        }

    @property
    def probe_polarization(self):
        return (self._obj.attrs.get('probe_polarization_theta'),
                self._obj.attrs.get('probe_polarization_alpha'))

    @property
    def pump_polarization(self):
        return (self._obj.attrs.get('pump_polarization_theta'),
                self._obj.attrs.get('pump_polarization_alpha'))

    @property
    def prebinning(self):
        prebinning = {}
        for d in self._obj.indexes:
            if '{}_prebinning'.format(d) in self._obj.attrs:
                prebinning[d] = self._obj.attrs['{}_prebinning'.format(d)]

        return prebinning

    @property
    def monochromator_info(self):
        return {
            'grating_lines_per_mm': self._obj.attrs.get('grating_lines_per_mm'),
        }

    @property
    def undulator_info(self):
        return {
            'gap': self._obj.attrs.get('undulator_gap'),
            'z': self._obj.attrs.get('undulator_z'),
            'harmonic': self._obj.attrs.get('undulator_harmonic'),
            'polarization': self._obj.attrs.get('undulator_polarization'),
            'type': self._obj.attrs.get('undulator_type'),
        }

    @property
    def analyzer_detail(self):
        return {
            'name': self._obj.attrs.get('analyzer_name'),
            'parallel_deflectors': self._obj.attrs.get('parallel_deflectors'),
            'perpendicular_deflectors': self._obj.attrs.get('perpendicular_deflectors'),
            'type': self._obj.attrs.get('analyzer_type'),
            'radius': self._obj.attrs.get('analyzer_radius'),
        }

    @property
    def temp(self):
        """
        TODO, agressively normalize attributes across different spectrometers
        :return:
        """
        prefered_attrs = ['TA', 'ta', 't_a', 'T_A', 'T_1', 't_1', 't1', 'T1', 'temp', 'temp_sample', 'temp_cryotip',
                          'temperature_sensor_b', 'temperature_sensor_a']
        for attr in prefered_attrs:
            if attr in self._obj.attrs:
                return float(self._obj.attrs[attr])

        raise AttributeError('Could not read temperature off any standard attr')

    @property
    def condensed_attrs(self):
        """
        Since we enforce camelcase on attributes, this is a reasonable filter that catches
        the ones we don't use very often.
        :return:
        """
        return {k: v for k, v in self._obj.attrs.items() if k[0].islower()}

    @property
    def referenced_scans(self):
        """
        Produces a dataframe which has all the scans which reference this one. This only makes sense for maps.
        :return:
        """

        if self.spectrum_type == 'map':
            df = self._obj.attrs['df']
            return df[(df.spectrum_type != 'map') & (df.ref_id == self._obj.id)]
        else:
            assert self.spectrum_type in {'ucut', 'spem'}
            df = self.df_until_type(spectrum_type=('ucut', 'spem',))
            return df

    def generic_fermi_surface(self, fermi_energy):
        return self.fat_sel(eV=fermi_energy)

    @property
    def fermi_surface(self):
        return self.fat_sel(eV=0)

    def __init__(self, xarray_obj):
        self._obj = xarray_obj


    @staticmethod
    def dict_to_html(d):
        return """
        <table>
          <thead>
            <tr>
              <th>Key</th>
              <th>Value</th>
            </tr>
          </thead>
          <tbody>
            {rows}
          </tbody>
        </table>
        """.format(
            rows=''.join(['<tr><td>{}</td><td>{}</td></tr>'.format(k, v) for k, v in d.items()])
        )

    def _repr_html_full_coords(self, coords):
        def coordinate_dataarray_to_flat_rep(value):
            if not isinstance(value, xr.DataArray):
                return value

            return '<span>{min:.3g}<strong> to </strong>{max:.3g}<strong> by </strong>{delta:.3g}</span>'.format(
                min=value.min().item(),
                max=value.max().item(),
                delta=value.values[1] - value.values[0],
            )

        return ARPESAccessorBase.dict_to_html({k: coordinate_dataarray_to_flat_rep(v) for k, v in coords.items()})

    def _repr_html_spectrometer_info(self):
        skip_keys = {'dof',}
        ordered_settings = OrderedDict(self.spectrometer_settings)
        ordered_settings.update({k: v for k, v in self.spectrometer.items()
                                 if k not in skip_keys})

        return ARPESAccessorBase.dict_to_html(ordered_settings)

    @staticmethod
    def _repr_html_experimental_conditions(conditions):
        transforms = {
            'polarization': lambda p: {
                'p': 'Linear Horizontal',
                's': 'Linear Vertical',
                'rc': 'Right Circular',
                'lc': 'Left Circular',
                's-p': 'Linear Dichroism',
                'p-s': 'Linear Dichroism',
                'rc-lc': 'Circular Dichroism',
                'lc-rc': 'Circular Dichroism',
            }.get(p, p),
            'hv': '{} eV'.format,
            'temp': '{} Kelvin'.format,
        }

        id = lambda x: x

        return ARPESAccessorBase.dict_to_html(
            {k: transforms.get(k, id)(v) for k, v in conditions.items() if v is not None})

    def _repr_html_(self):
        skip_data_vars = {'time',}

        if isinstance(self._obj, xr.Dataset):
            to_plot = [k for k in self._obj.data_vars.keys() if k not in skip_data_vars]
            to_plot = [k for k in to_plot if 1 <= len(self._obj[k].dims) < 3]
            to_plot = to_plot[:5]

            if to_plot:
                _, ax = plt.subplots(1, len(to_plot), figsize=(len(to_plot) * 3, 3,))
                if len(to_plot) == 1:
                    ax = [ax]

                for i, plot_var in enumerate(to_plot):
                    self._obj[plot_var].plot(ax=ax[i])
                    fancy_labels(ax[i])
                    ax[i].set_title(plot_var.replace('_', ' '))

                remove_colorbars()

        else:
            if 1 <= len(self._obj.dims) < 3:
                fig, ax = plt.subplots(1, 1, figsize=(4, 3))
                self._obj.plot(ax=ax)
                fancy_labels(ax)
                ax.set_title('')

                remove_colorbars()

        wrapper_style = 'style="display: flex; flex-direction: row;"'

        try:
            name = self.df_index
        except:
            if 'id' in self._obj.attrs:
                name = 'ID: ' + str(self._obj.attrs['id'])[:9] + '...'
            else:
                name = 'No name'

        warning = ''

        if len(self._obj.attrs) < 10:
            warning = ':  <span style="color: red;">Few Attributes, Data Is Summed?</span>'

        return """
        <header><strong>{name}{warning}</strong></header>
        <div {wrapper_style}>
        <details open>
            <summary>Experimental Conditions</summary>
            {conditions}
        </details>
        <details open>
            <summary>Full Coordinates</summary>
            {coordinates}
        </details>
        <details open>
            <summary>Spectrometer</summary>
            {spectrometer_info}
        </details>
        </div>
        """.format(
            name=name,
            warning=warning,
            wrapper_style=wrapper_style,
            conditions=self._repr_html_experimental_conditions(self.experimental_conditions),
            coordinates=self._repr_html_full_coords({k: v for k, v in self.full_coords.items() if v is not None}),
            spectrometer_info=self._repr_html_spectrometer_info(),
        )


@xr.register_dataarray_accessor('S')
class ARPESDataArrayAccessor(ARPESAccessorBase):
    def plot(self, *args, **kwargs):
        with plt.rc_context(rc={'text.usetex': False}):
            self._obj.plot(*args, **kwargs)

    def show(self, **kwargs):
        image_tool = ImageTool(**kwargs)
        return image_tool.make_tool(self._obj)

    def show_d2(self, **kwargs):
        curve_tool = CurvatureTool(**kwargs)
        return curve_tool.make_tool(self._obj)

    def show_band_tool(self, **kwargs):
        band_tool = BandTool(**kwargs)
        return band_tool.make_tool(self._obj)

    def fs_plot(self, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_fs'.format(self.label))
            kwargs['out'] = out
        return plotting.labeled_fermi_surface(self._obj, **kwargs)

    def cut_dispersion_plot(self, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_cut_dispersion'.format(self.label))
            kwargs['out'] = out
        return plotting.cut_dispersion_plot(self._obj, **kwargs)

    def dispersion_plot(self, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_dispersion'.format(self.label))
            kwargs['out'] = out
        return plotting.fancy_dispersion(self._obj, **kwargs)

    def isosurface_plot(self, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_isosurface'.format(self.label))
            kwargs['out'] = out
        return plotting.plot_isosurface(self._obj, **kwargs)


    def subtraction_reference_plot(self, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_subtraction_reference'.format(self.label))
            kwargs['out'] = out

        return plotting.tarpes.plot_subtraction_reference(self._obj, **kwargs)

    def fermi_edge_reference_plot(self, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_fermi_edge_reference'.format(self.label))
            kwargs['out'] = out

        return plotting.fermi_edge.fermi_edge_reference(self._obj, **kwargs)


    def _referenced_scans_for_spatial_plot(self, use_id=True, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        label = self._obj.attrs['id'] if use_id else self.label
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_reference_scan_fs'.format(label))
            kwargs['out'] = out

        return plotting.reference_scan_spatial(self._obj, **kwargs)


    def _referenced_scans_for_map_plot(self, use_id=True, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        label = self._obj.attrs['id'] if use_id else self.label
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_reference_scan_fs'.format(label))
            kwargs['out'] = out

        return plotting.reference_scan_fermi_surface(self._obj, **kwargs)

    def _referenced_scans_for_hv_map_plot(self, use_id=True, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        label = self._obj.attrs['id'] if use_id else self.label
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_hv_reference_scan'.format(label))
            out = '{}_hv_reference_scan.png'.format(label)
            kwargs['out'] = out

        return plotting.hv_reference_scan(self._obj, **kwargs)

    def _simple_spectrum_reference_plot(self, use_id=True, pattern='{}.png', **kwargs):
        out = kwargs.get('out')
        label = self._obj.attrs['id'] if use_id else self.label
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_spectrum_reference'.format(label))
            kwargs['out'] = out

        return plotting.fancy_dispersion(self._obj, **kwargs)

    def cut_nan_coords(self):
        slices = dict()
        for cname, cvalue in self._obj.coords.items():
            try:
                end_ind = np.where(np.isnan(cvalue.values))[0][0]
                end_ind = None if end_ind == -1 else end_ind
                slices[cname] = slice(None, end_ind)
            except IndexError:
                pass

        return self._obj.isel(**slices)

    def nan_to_num(self, x=0):
        """
        xarray version of numpy.nan_to_num
        :param x:
        :return:
        """

        data = self._obj.copy(deep=True)
        assert isinstance(data, xr.DataArray)
        data.values[np.isnan(data.values)] = x
        return data

    def reference_plot(self, **kwargs):
        if self.spectrum_type == 'map':
            return self._referenced_scans_for_map_plot(**kwargs)
        elif self.spectrum_type == 'hv_map':
            return self._referenced_scans_for_hv_map_plot(**kwargs)
        elif self.spectrum_type == 'spectrum':
            return self._simple_spectrum_reference_plot(**kwargs)
        elif self.spectrum_type in {'ucut', 'spem'}:
            return self._referenced_scans_for_spatial_plot(**kwargs)
        elif self.spectrum_type in {'cut'}:
            return None
        else:
            import pdb
            pdb.set_trace()

    @property
    def eV(self):
        eV = None
        if 'eV' in self._obj.dims:
            eV = self._obj.eV
        else:
            for dof in set(self._obj.dims):
                if 'eV' in dof:
                    eV = self._obj[dof]
        return eV


NORMALIZED_DIM_NAMES = ['x', 'y', 'z', 'w']

@xr.register_dataset_accessor('T')
@xr.register_dataarray_accessor('T')
class GenericAccessorTools:
    _obj = None

    def extent(self, *args, dims=None):
        """
        Returns an "extent" array that can be used to draw with plt.imshow
        :return:
        """

        if dims is None:
            if not args:
                dims = self._obj.dims
            else:
                dims = args

        assert len(dims) == 2 and 'You must supply exactly two dims to `.T.extent` not {}'.format(dims)
        return [
            self._obj.coords[dims[0]][0].item(),
            self._obj.coords[dims[0]][-1].item(),
            self._obj.coords[dims[1]][0].item(),
            self._obj.coords[dims[1]][-1].item(),
        ]

    def drop_nan(self):
        assert len(self._obj.dims) == 1

        mask = np.logical_not(np.isnan(self._obj.values))
        return self._obj.isel(**dict([[self._obj.dims[0], mask]]))

    def shift_coords(self, dims, shift):
        if not isinstance(shift, np.ndarray):
            shift = np.ones((len(dims),)) * shift

        def transform(data):
            new_shift = shift
            for _ in range(len(dims)):
                new_shift = np.expand_dims(new_shift, 0)

            return data + new_shift

        return self.transform_coords(dims, transform)

    def scale_coords(self, dims, scale):
        if not isinstance(scale, np.ndarray):
            n_dims = len(dims)
            scale = np.identity(n_dims) * scale
        elif len(scale.shape) == 1:
            scale = np.diag(scale)

        return self.transform_coords(dims, scale)

    def transform_coords(self, dims, transform):
        """
        Transforms the given coordinate values according to transform, should either be a function
        from a len(dims) x size of raveled coordinate array to len(dims) x size of raveled_coordinate
        array or a linear transformation as a matrix which is multiplied into such an array.
        :param dims: List of dimensions that should be transformed
        :param transform: The transformation to apply, can either be a function, or a matrix
        :return:
        """
        as_array = np.stack([self._obj.data_vars[d].values for d in dims], axis=-1)

        if isinstance(transform, np.ndarray):
            transformed = np.dot(as_array, transform)
        else:
            transformed = transform(as_array)

        copied = self._obj.copy(deep=True)

        for d, arr in zip(dims, np.split(transformed, transformed.shape[-1], axis=-1)):
            copied.data_vars[d].values = np.squeeze(arr, axis=-1)

        return copied

    def filter_vars(self, f):
        return xr.Dataset(data_vars={
            k: v for k, v in self._obj.data_vars.items() if f(v, k)
        }, attrs=self._obj.attrs)

    def var_startswith(self, fragment):
        return self.filter_vars(lambda _, k: k.startswith(fragment))

    def var_contains(self, fragment):
        """
        Filters a dataset's variables based on whether the name contains the fragment
        :param fragment:
        :return:
        """
        return self.filter_vars(lambda _, k: fragment in k)

    def coordinatize(self, as_coordinate_name):
        """
        Remarkably, `coordinatize` is a word

        :return:
        """
        assert len(self._obj.dims) == 1

        d = self._obj.dims[0]
        o = self._obj.rename(dict([[d, as_coordinate_name]])).copy(deep=True)
        o.coords[as_coordinate_name] = o.values

        return o

    def ravel(self, as_dataset=False):
        """
        Converts to a flat representation where the coordinate values are also present.
        Extremely valuable for plotting a dataset with coordinates, X, Y and values Z(X,Y)
        on a scatter plot in 3D.

        By default the data is listed under the key 'data'.

        :return:
        """

        assert isinstance(self._obj, xr.DataArray)

        dims = self._obj.dims
        coords_as_list = [self._obj.coords[d].values for d in dims]
        raveled_coordinates = dict(zip(dims, [cs.ravel() for cs in np.meshgrid(*coords_as_list)]))
        assert 'data' not in raveled_coordinates
        raveled_coordinates['data'] = self._obj.values.ravel()

        return raveled_coordinates

    def meshgrid(self, as_dataset=False):
        assert isinstance(self._obj, xr.DataArray)

        dims = self._obj.dims
        coords_as_list = [self._obj.coords[d].values for d in dims]
        meshed_coordinates = dict(zip(dims, [cs for cs in np.meshgrid(*coords_as_list)]))
        assert 'data' not in meshed_coordinates
        meshed_coordinates['data'] = self._obj.values

        if as_dataset:
            # this could use a bit of cleaning up
            faked = ['x', 'y', 'z', 'w']
            meshed_coordinates = {k: (faked[:len(v.shape)], v) for k, v in meshed_coordinates.items() if k != 'data'}

            return xr.Dataset(meshed_coordinates)

        return meshed_coordinates

    def to_arrays(self):
        """
        Useful for rapidly converting into a format than can be `plt.scatter`ed
        or similar.

        Ex:

        plt.scatter(*data.T.as_arrays(), marker='s')
        :return:
        """
        assert len(self._obj.dims) == 1

        return [self._obj.coords[self._obj.dims[0]].values, self._obj.values]

    def clean_outliers(self, clip=0.5):
        low, high = np.percentile(self._obj.values, [clip, 100 - clip])
        copied = self._obj.copy(deep=True)
        copied.values[copied.values < low] = low
        copied.values[copied.values > high] = high
        return copied

    def as_movie(self, time_dim=None, pattern='{}.png', **kwargs):
        if time_dim is None:
            time_dim = self._obj.dims[-1]

        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = pattern.format('{}_animation'.format(self._obj.S.label))
            kwargs['out'] = out
        return plotting.plot_movie(self._obj, time_dim, **kwargs)

    def filter_coord(self, coordinate_name, sieve):
        """
        Filters a dataset along a coordinate.

        Sieve should be a function which accepts a coordinate value and the slice
        of the data along that dimension.

        :param coordinate_name:
        :param sieve:
        :return:
        """
        mask = np.array([i for i, c in enumerate(self._obj.coords[coordinate_name])
                         if sieve(c, self._obj.isel(**dict([[coordinate_name, i]])))])
        return self._obj.isel(**dict([[coordinate_name, mask]]))

    def iterate_axis(self, axis_name_or_axes):
        if isinstance(axis_name_or_axes, int):
            axis_name_or_axes = self._obj.dims[axis_name_or_axes]

        if isinstance(axis_name_or_axes, str):
            axis_name_or_axes = [axis_name_or_axes]

        coord_iterators = [self._obj.coords[d].values for d in axis_name_or_axes]
        for indices in itertools.product(*[range(len(c)) for c in coord_iterators]):
            cut_coords = [cs[index] for cs, index in zip(coord_iterators, indices)]
            coords_dict = dict(zip(axis_name_or_axes, cut_coords))
            yield coords_dict, self._obj.sel(method='nearest', **coords_dict)

    def map_axes(self, axes, fn, dtype=None, **kwargs):
        if isinstance(self._obj, xr.Dataset):
            raise TypeError('map_axes can only work on xr.DataArrays for now because of '
                            'how the type inference works')
        obj = self._obj.copy(deep=True)

        if dtype is not None:
            obj.values = np.ndarray(shape=obj.values.shape, dtype=dtype)

        type_assigned = False
        for coord, value in self.iterate_axis(axes):
            new_value = fn(value, coord)

            if dtype is None:
                if not type_assigned:
                    obj.values = np.ndarray(shape=obj.values.shape, dtype=new_value.data.dtype)
                    type_assigned = True

                obj.loc[coord] = new_value.values
            else:
                obj.loc[coord] = new_value

        return obj

    def map(self, fn, **kwargs):
        return apply_dataarray(self._obj, np.vectorize(fn, **kwargs))

    def enumerate_iter_coords(self):
        coords_list = [self._obj.coords[d].values for d in self._obj.dims]
        for indices in itertools.product(*[range(len(c)) for c in coords_list]):
            cut_coords = [cs[index] for cs, index in zip(coords_list, indices)]
            yield indices, dict(zip(self._obj.dims, cut_coords))

    def iter_coords(self, dim_names=None):
        if dim_names is None:
            dim_names = self._obj.dims
        for ts in itertools.product(*[self._obj.coords[d].values for d in self._obj.dims]):
            yield dict(zip(self._obj.dims, ts))

    def range(self, generic_dim_names=True):
        indexed_coords = [self._obj.coords[d] for d in self._obj.dims]
        indexed_ranges = [(np.min(coord.values), np.max(coord.values)) for coord in indexed_coords]

        dim_names = self._obj.dims
        if generic_dim_names:
            dim_names = NORMALIZED_DIM_NAMES[:len(dim_names)]

        return dict(zip(dim_names, indexed_ranges))

    def stride(self, *args, generic_dim_names=True):
        indexed_coords = [self._obj.coords[d] for d in self._obj.dims]
        indexed_strides = [coord.values[1] - coord.values[0] for coord in indexed_coords]

        dim_names = self._obj.dims
        if generic_dim_names:
            dim_names = NORMALIZED_DIM_NAMES[:len(dim_names)]

        result = dict(zip(dim_names, indexed_strides))

        if args:
            if len(args) == 1:
                if not isinstance(args[0], str):
                    # if passed list of strs as argument
                    result = [result[selected_names] for selected_names in args[0]]
                else:
                    # if passed single name as argument
                    result = result[args[0]]
            else:
                # if passed several names as arguments
                result = [result[selected_names] for selected_names in args]

        return result

    def shift_by(self, other, shift_axis=None, zero_nans=True, shift_coords=False):
        # for now we only support shifting by a one dimensional array

        data = self._obj

        assert len(other.dims) == 1

        if shift_coords:
            mean_shift = np.mean(other)
            other -= mean_shift

        by_axis = other.dims[0]
        if shift_axis is None:
            option_dims = list(data.dims)
            option_dims.remove(by_axis)
            assert len(option_dims) == 1
            shift_axis = option_dims[0]

        shift_amount = -other.values / data.T.stride(generic_dim_names=False)[shift_axis]

        shifted_data = shift_by(data.values, shift_amount,
                                axis=list(data.dims).index(shift_axis),
                                by_axis=list(data.dims).index(by_axis), order=1)

        if zero_nans:
            shifted_data[np.isnan(shifted_data)] = 0

        coords = copy.deepcopy(data.coords)
        if shift_coords:
            coords[shift_axis] -= mean_shift

        return xr.DataArray(
            shifted_data,
            coords,
            data.dims,
            attrs=data.attrs.copy(),
        )


    def __init__(self, xarray_obj: DataType):
        self._obj = xarray_obj


@xr.register_dataset_accessor('F')
class ARPESDatasetFitToolAccessor:
    _obj = None

    def __init__(self, xarray_obj: DataType):
        self._obj = xarray_obj

    def eval(self, *args, **kwargs):
        return self._obj.results.T.map(lambda x: x.eval(*args, **kwargs))

    def show(self):
        fit_diagnostic_tool = FitCheckTool()
        return fit_diagnostic_tool.make_tool(self._obj)

    def p(self, param_name):
        return self._obj.results.F.p(param_name)

    def s(self, param_name):
        return self._obj.results.F.s(param_name)

    def plot_param(self, param_name, **kwargs):
        return self._obj.results.F.plot_param(param_name, **kwargs)


@xr.register_dataarray_accessor('F')
class ARPESFitToolsAccessor:
    _obj = None

    def __init__(self, xarray_obj: DataType):
        self._obj = xarray_obj

    def plot_param(self, param_name, **kwargs):
        plotting.plot_parameter(self._obj, param_name, **kwargs)

    def param_as_dataset(self, param_name):
        return xr.Dataset({
            'value': self.p(param_name),
            'error': self.s(param_name),
        })

    def show(self):
        fit_diagnostic_tool = FitCheckTool()
        return fit_diagnostic_tool.make_tool(self._obj)

    def p(self, param_name):
        return self._obj.T.map(param_getter(param_name), otypes=[np.float])

    def s(self, param_name):
        return self._obj.T.map(param_stderr_getter(param_name), otypes=[np.float])

    @property
    def bands(self):
        """
        This should probably instantiate appropriate types
        :return:
        """
        band_names = self.band_names

        bands = {l: MultifitBand(label=l, data=self._obj) for l in band_names}

        return bands

    @property
    def band_names(self):
        collected_band_names = set()

        for item in self._obj.values.ravel():
            if item is None:
                continue

            band_names = [k[:-6] for k in item.params.keys() if 'center' in k]
            collected_band_names = collected_band_names.union(set(band_names))

        return collected_band_names

    @property
    def parameter_names(self):
        collected_parameter_names = set()

        for item in self._obj.values.ravel():
            if item is None:
                continue

            param_names = [k for k in item.params.keys()]
            collected_parameter_names = collected_parameter_names.union(set(param_names))

        return collected_parameter_names


    def show_fit_diagnostic(self):
        """
        alias for ``.show``
        :return:
        """
        return self.show()


@xr.register_dataset_accessor('S')
class ARPESDatasetAccessor(ARPESAccessorBase):
    def __getattr__(self, item):
        return getattr(self._obj.S.spectrum.S, item)

    def polarization_plot(self, **kwargs):
        out = kwargs.get('out')
        if out is not None and isinstance(out, bool):
            out = '{}_spin_polarization.png'.format(self.label)
            kwargs['out'] = out
        return plotting.spin_polarized_spectrum(self._obj, **kwargs)

    @property
    def is_spatial(self):
        try:
            return self.spectrum.S.is_spatial
        except Exception as e:
            return False

    @property
    def spectrum(self) -> Optional[xr.DataArray]:
        spectrum = None
        if 'spectrum' in self._obj.data_vars:
            spectrum = self._obj.spectrum
        elif 'raw' in self._obj.data_vars:
            spectrum = self._obj.raw
        elif '__xarray_dataarray_variable__' in self._obj.data_vars:
            spectrum = self._obj.__xarray_dataarray_variable__
        elif any('spectrum' in dv for dv in self._obj.data_vars):
            spectrum = self._obj[list(self._obj.data_vars)[list('spectrum' in dv for dv in self._obj.data_vars).index(True)]]

        if spectrum is not None and 'df' not in spectrum.attrs:
            spectrum.attrs['df'] = self._obj.attrs.get('df', None)

        return spectrum

    @property
    def spectra(self):
        spectra = []
        for dv in list(self._obj.data_vars):
            if 'spectrum' in dv:
                spectra.append(self._obj[dv])

        return spectra

    @property
    def is_multi_region(self):
        return len(self.spectra) > 1

    @property
    def spectrum_type(self):
        try:
            # this isn't the smartest thing in the world,
            # but it should allow some old code to keep working on datasets transparently
            return self.spectrum.S.spectrum_type
        except Exception:
            return 'dataset'

    @property
    def degrees_of_freedom(self):
        return set(self.spectrum.dims)

    @property
    def spectrum_degrees_of_freedom(self):
        return self.degrees_of_freedom.intersection({
            'eV', 'phi', 'pixel', 'kx', 'kp'
        })

    @property
    def scan_degrees_of_freedom(self):
        return self.degrees_of_freedom.difference(self.spectrum_degrees_of_freedom)

    def reference_plot(self, **kwargs):
        """
        A bit of a misnomer because this actually makes many plots. For full datasets, the relevant components
        are:

        1. Temperature as function of scan DOF
        2. Photocurrent as a function of scan DOF
        3. Photocurrent normalized + unnormalized figures, in particular
           i. The reference plots for the photocurrent normalized spectrum
           ii. The normalized total cycle intensity over scan DoF, i.e. cycle vs scan DOF integrated over E, phi
           iii. For delay scans:
             1. Fermi location as a function of scan DoF, integrated over phi
             2. Subtraction scans
        4. For spatial scans:
           i. energy/angle integrated spatial maps with subsequent measurements indicated
           2. energy/angle integrated FS spatial maps with subsequent measurements indicated

        :param kwargs:
        :return:
        """
        scan_dofs_integrated = self._obj.sum(*list(self.scan_degrees_of_freedom))
        original_out = kwargs.get('out')

        # make figures for temperature, photocurrent, delay
        make_figures_for = ['T', 'IG_nA', 'current', 'photocurrent']
        name_normalization = {
            'T': 'T',
            'IG_nA': 'photocurrent',
            'current': 'photocurrent'
        }

        for figure_item in make_figures_for:
            if figure_item not in self._obj.data_vars:
                continue

            name = name_normalization.get(figure_item, figure_item)
            data_var = self._obj[figure_item]
            out = '{}_{}_spec_integrated_reference.png'.format(self.label, name)
            return plotting.scan_var_reference_plot(
                data_var, title='Reference {}'.format(name), out=out)

        # may also want to make reference figures summing over cycle, or summing over beta

        # make photocurrent normalized figures
        try:
            normalized = self._obj / self._obj.IG_nA
            normalized.S.make_spectrum_reference_plots(prefix='norm_PC_', out=True)
        except:
            pass

        self.make_spectrum_reference_plots(out=True)

    def make_spectrum_reference_plots(self, prefix='', **kwargs):
        """
        Photocurrent normalized + unnormalized figures, in particular:

        i. The reference plots for the photocurrent normalized spectrum
        ii. The normalized total cycle intensity over scan DoF, i.e. cycle vs scan DOF integrated over E, phi
        iii. For delay scans:
          1. Fermi location as a function of scan DoF, integrated over phi
          2. Subtraction scans
        """
        self.spectrum.S.reference_plot(pattern=prefix + '{}.png', **kwargs)

        if self.is_spatial:
            referenced = self.referenced_scans

        if 'cycle' in self._obj.coords:
            integrated_over_scan = self._obj.sum(*list(self.spectrum_degrees_of_freedom))
            integrated_over_scan.S.spectrum.S.reference_plot(pattern=prefix + 'sum_spec_DoF_{}.png', **kwargs)

        if 'delay' in self._obj.coords:
            dims = self.spectrum_degrees_of_freedom
            dims.remove('eV')
            angle_integrated = self._obj.sum(*list(dims))

            # subtraction scan
            self.spectrum.S.subtraction_reference_plots(pattern=prefix + '{}.png', **kwargs)
            angle_integrated.S.fermi_edge_reference_plots(pattern=prefix + '{}.png', **kwargs)

    def __init__(self, xarray_obj):
        super().__init__(xarray_obj)
        self._spectrum = None



import numpy as np

import arpes.xarray_extensions # pylint: disable=unused-import
import xarray as xr
from arpes.endstations import FITSEndstation, HemisphericalEndstation

__all__ = ('ALGMainChamber',)


class ALGMainChamber(HemisphericalEndstation, FITSEndstation):
    PRINCIPAL_NAME = 'ALG-Main'
    ALIASES = ['MC', 'ALG-Main', 'ALG-MC', 'ALG-Hemisphere', 'ALG-Main Chamber',]

    ATTR_TRANSFORMS = {
        'START_T': lambda l: {'time': ' '.join(l.split(' ')[1:]).lower(),
                              'date': l.split(' ')[0]},
    }

    RENAME_KEYS = {
        'Phi': 'chi',
        'Beta': 'beta',
        'Theta': 'theta',
        'Azimuth': 'chi',
        'Alpha': 'alpha',
        'Pump_energy_uJcm2': 'pump_fluence',
        'T0_ps': 't0_nominal',
        'W_func': 'workfunction',
        'Slit': 'slit',
        'LMOTOR0': 'x',
        'LMOTOR1': 'y',
        'LMOTOR2': 'z',
        'LMOTOR3': 'theta',
        'LMOTOR4': 'beta',
        'LMOTOR5': 'chi',
        'LMOTOR6': 'delay',

        'SFLNM0': 'lens_mode_name',
        'SFFR_0': 'frames_per_slice',

        'SFBA_0': 'phi_prebinning',
        'SFBE0': 'eV_prebinning',
    }

    MERGE_ATTRS = {
        'analyzer': 'Specs PHOIBOS 150',
        'analyzer_name': 'Specs PHOIBOS 150',
        'parallel_deflectors': False,
        'perpendicular_deflectors': False,
        'analyzer_radius': 150,
        'analyzer_type': 'hemispherical',
        'mcp_voltage': None,
        'probe_linewidth': 0.015,
    }

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict=None):
        data.attrs['hv'] = 5.93
        data.attrs['alpha'] = 0
        data.attrs['psi'] = 0

        # by default we use this value since this isnear the center of the spectrometer window
        data.attrs['phi_offset'] = 0.405
        for spectrum in data.S.spectra:
            spectrum.attrs['hv'] = 5.93  # only photon energy available on this chamber
            spectrum.attrs['alpha'] = 0
            spectrum.attrs['psi'] = 0
            spectrum.attrs['phi_offset'] = 0.405

        data = super().postprocess_final(data, scan_desc)

        if 'beta' in data.coords:
            data.coords['beta'].values = data.coords['beta'].values * np.pi / 180

        return data



import functools
from ast import literal_eval
from collections import Iterable

import numpy as np
from astropy.io.fits.hdu.table import BinTableHDU
from numpy import ndarray

from arpes.utilities.funcutils import collect_leaves, iter_leaves
from typing import Any, Dict, List, Optional, Tuple, Union

__all__ = ('extract_coords', 'find_clean_coords',)

DEFAULT_DIMENSION_RENAMINGS = {
    'Beta': 'beta',
    'Theta': 'theta',
    'Delay': 'delay',
    'Sample-X': 'cycle',
    'null': 'cycle',
    'Mira': 'pump_power',
    'X': 'x',
    'Y': 'y',
    'Z': 'z',
}

CoordsDict = Dict[str, ndarray]
Dimension = str


def extract_coords(attrs: Dict[str, Any], dimension_renamings: Dict[str, str] = None) \
        -> Tuple[CoordsDict, List[Dimension], List[int]]:
    """
    Does the hard work of extracting coordinates from the scan description.
    :param attrs:
    :param dimension_renamings:
    :return:
    """
    if dimension_renamings is None:
        dimension_renamings = DEFAULT_DIMENSION_RENAMINGS

    try:
        n_loops = attrs['LWLVLPN']
    except KeyError:
        # Looks like no scan, this happens for instance in the SToF when you take a single
        # EDC
        return {}, [], (),
    scan_dimension = []
    scan_shape = []
    scan_coords = {}
    for loop in range(n_loops):
        n_scan_dimensions = attrs['NMSBDV%g' % loop]
        if attrs['SCNTYP%g' % loop] == 0:  # computed
            for i in range(n_scan_dimensions):
                name, start, end, n = (
                    attrs['NM_%g_%g' % (loop, i,)],
                    # attrs['UN_0_%g' % i],
                    float(attrs['ST_%g_%g' % (loop, i,)]),
                    float(attrs['EN_%g_%g' % (loop, i,)]),
                    int(attrs['N_%g_%g' % (loop, i,)]),
                )

                name = dimension_renamings.get(name, name)

                scan_dimension.append(name)
                scan_shape.append(n)
                scan_coords[name] = np.linspace(start, end, n, endpoint=True)
        else:  # tabulated scan, this is more complicated
            name, n = (
                attrs['NM_%g_0' % loop],
                attrs['NMPOS_%g' % loop],
            )

            try:
                n_regions_key = {'Delay': 'DS_NR'}.get(name, 'DS_NR')
                n_regions = attrs[n_regions_key]

                name = dimension_renamings.get(name, name)
            except KeyError:
                if 'ST_{}_1'.format(loop) in attrs:
                    assert(False and "More than one region detected but unhandled.")

                n_regions = 1
                name = dimension_renamings.get(name, name)

            coord = np.array(())
            for region in range(n_regions):
                start, end, n = (
                    attrs['ST_%g_%g' % (loop, region,)],
                    attrs['EN_%g_%g' % (loop, region,)],
                    attrs['N_%g_%g' % (loop, region,)],
                )

                coord = np.concatenate((coord, np.linspace(
                    start, end, n, endpoint=True),))

            scan_dimension.append(name)
            scan_shape.append(len(coord))
            scan_coords[name] = coord
    return scan_coords, scan_dimension, scan_shape


def find_clean_coords(hdu: BinTableHDU, attrs: Dict[str, Any], spectra: Optional[Any] = None, mode: str = 'ToF',
                      dimension_renamings: Optional[Any] = None) \
        -> Tuple[CoordsDict, Dict[str, List[Dimension]], Dict[str, Any]]:
    """
    Determines the scan degrees of freedom, the shape of the actual "spectrum"
    and reads and parses the coordinates from the header information in the recorded
    scan.

    Note: because different scan configurations can have different values of the detector coordinates, such as
    for instance when you record in two different angular modes of the spectrometer or when you record XPS spectra
    at different binding energies, we need to be able to provide separate coordinates for each of the different scans.

    In the case where there is a unique coordinate, we will return only that coordinate, under the anticipated name,
    such as 'eV'. Otherwise, we will return the coordinates that different between the scan configurations under the
    spectrum name, and with unique names, such as 'eV-Swept_Spectra0'.

    TODO Write data loading tests to ensure we don't break MC compatibility

    :param spectra:
    :param hdu:
    :param attrs:
    :param mode: Available modes are "ToF", "MC". This customizes the read process
    :param dimension_renamings:
    :return: (coordinates, dimensions, np shape of actual spectrum)
    """
    if dimension_renamings is None:
        dimension_renamings = DEFAULT_DIMENSION_RENAMINGS

    scan_coords, scan_dimension, scan_shape = extract_coords(attrs, dimension_renamings=dimension_renamings)

    # bit of a hack to deal with the internal motor used for the swept spectra being considered as a cycle
    if 'cycle' in scan_coords and len(scan_coords['cycle']) > 200:
        idx = scan_dimension.index('cycle')

        real_data_for_cycle = hdu.data.columns['null'].array

        scan_coords['cycle'] = real_data_for_cycle
        scan_shape[idx] = len(real_data_for_cycle)

    scan_dimension = [dimension_renamings.get(s, s) for s in scan_dimension[::-1]]
    scan_coords = {dimension_renamings.get(k, k): v for k, v in scan_coords.items()}
    extra_coords = {}
    scan_shape = scan_shape[::-1]

    spectrum_shapes = {}
    dimensions_for_spectra = {}

    if spectra is None:
        spectra = hdu.columns.names

    if isinstance(spectra, str):
        spectra = [spectra]

    for spectrum_key in spectra:
        skip_names = {
            lambda name: True if ('beamview' in name or 'IMAQdx' in name) else False,
        }

        if spectrum_key is None:
            spectrum_key = hdu.columns.names[-1]

        if isinstance(spectrum_key, str):
            spectrum_key = hdu.columns.names.index(spectrum_key) + 1

        spectrum_name = hdu.columns.names[spectrum_key - 1]
        loaded_shape_from_header = False
        desc = None

        should_skip = False
        for skipped in skip_names:
            if callable(skipped) and skipped(spectrum_name):
                should_skip = True
            elif skipped == spectrum_name:
                should_skip = True
        if should_skip:
            continue

        try:
            offset = hdu.header['TRVAL%g' % spectrum_key]
            delta = hdu.header['TDELT%g' % spectrum_key]
            offset = literal_eval(offset) if isinstance(offset, str) else offset
            delta = literal_eval(delta) if isinstance(delta, str) else delta

            try:
                shape = hdu.header['TDIM%g' % spectrum_key]
                shape = literal_eval(shape) if isinstance(shape, str) else shape
                loaded_shape_from_header = True
            except:
                shape = hdu.data.field(spectrum_key - 1).shape

            try:
                desc = hdu.header['TDESC%g' % spectrum_key]
                if '(' in desc:
                    # might be a malformed tuple, we can't use literal_eval unfortunately
                    desc = desc.replace('(', '').replace(')', '').split(',')

                if isinstance(desc, str):
                    desc = (desc,)
            except KeyError:
                pass

            if not isinstance(delta, Iterable):
                delta = (delta,)
            if not isinstance(offset, Iterable):
                offset = (offset,)

        except KeyError:
            # if TRVAL{spectrum_key} was not found this means that this column is scalar,
            # i.e. it has only one value at any point in the scan
            spectrum_shapes[spectrum_name] = scan_shape
            dimensions_for_spectra[spectrum_name] = scan_dimension
            continue

        if not scan_shape and shape[0] == 1:
            # the ToF pads with ones on single EDCs
            shape = shape[1:]

        if mode == 'ToF':
            rest_shape = shape[len(scan_shape):]
        else:
            if isinstance(desc, tuple):
                rest_shape = shape[-len(desc):]
            elif not loaded_shape_from_header:
                rest_shape = shape[1:]
            else:
                rest_shape = shape

        assert(len(offset) == len(delta) and len(delta) == len(rest_shape))

        # Build the actually coordinates
        coords = [np.linspace(o, o + s * d, s, endpoint=False)
                  for o, d, s in zip(offset, delta, rest_shape)]

        # We need to do smarter inference here
        def infer_hemisphere_dimensions() -> List[Dimension]:
            # scans can be two dimensional per frame, or a
            # scan can be either E or K integrated, or something I've never seen before
            # try to get the description or the UNIT
            if desc is not None:
                RECOGNIZED_DESCRIPTIONS = {
                    'eV': 'eV',
                    'pixels': 'pixel',
                    'pixel': 'pixel',
                }

                if all(d in RECOGNIZED_DESCRIPTIONS for d in desc):
                    return [RECOGNIZED_DESCRIPTIONS[d] for d in desc]

            try:
                # TODO read above like desc
                unit = hdu.header['TUNIT{}'.format(spectrum_key)]
                RECOGNIZED_UNITS = {
                    # it's probably 'arb' which doesn't tell us anything...
                    # because all spectra have arbitrary absolute intensity
                }
                if all(u in RECOGNIZED_UNITS for u in unit):
                    return [RECOGNIZED_UNITS[u] for u in unit]
            except KeyError:
                pass

            # Need to fall back on some human in the loop to improve the read
            # here
            import pdb
            pdb.set_trace()


        # TODO for cleanup in future, these should be provided by the implementing endstation class, so they do not
        # get so cluttered, best way will be to make this function a class method, and use class attributes for
        # each of `coord_names_for_spectrum`, etc. For now, patching to avoid error with the microscope camera images
        # at BL7
        coord_names_for_spectrum = {
            'Time_Spectra': ['time'],
            'Energy_Spectra': ['eV'],
            # MC hemisphere image, this can still be k-integrated, E-integrated, etc
            'wave':  ['time'],
            'targetPlus': ['time'],
            'targetMinus': ['time'],
            'Energy_Target_Up': ['eV'],
            'Energy_Target_Down': ['eV'],
            'Energy_Up': ['eV'],
            'Energy_Down': ['eV'],
            'Energy_Pol': ['eV'],
        }

        spectra_types = {
            'Fixed_Spectra',
            'Swept_Spectra',
        }

        time_spectra_type = {
            'Time_Target',
        }
        coord_names = None
        if spectrum_name not in coord_names_for_spectrum:
            # Don't remember what the MC ones were, so I will wait to do those again
            # Might have to add new items for new spectrometers as well
            if any(s in spectrum_name for s in spectra_types):
                coord_names = infer_hemisphere_dimensions
            elif any(s in spectrum_name for s in time_spectra_type):
                coord_names = ['time',]
            else:
                import pdb
                pdb.set_trace()
        else:
            coord_names = coord_names_for_spectrum[spectrum_name]

        if callable(coord_names):
            coord_names = coord_names()
            if len(coord_names) > 1 and mode == 'MC':
                # for whatever reason, the main chamber records data
                # in nonstandard byte order
                coord_names = coord_names[::-1]
                rest_shape = list(rest_shape)[::-1]
                coords = coords[::-1]

        coords_for_spectrum = dict(zip(coord_names, coords))
        # we need to store the coordinates that were kept in a table separately, because they are allowed to differ
        # between different scan configurations in the same file
        if mode == 'ToF':
            extra_coords.update(coords_for_spectrum)
        else:
            extra_coords[spectrum_name] = coords_for_spectrum
        dimensions_for_spectra[spectrum_name] = \
            tuple(scan_dimension) + tuple(coord_names)
        spectrum_shapes[spectrum_name] = tuple(scan_shape) + tuple(rest_shape)
        coords_for_spectrum.update(scan_coords)

    extra_coords.update(scan_coords)

    if mode != 'ToF':
        detector_coord_names = [k for k, v in extra_coords.items() if isinstance(v, dict)]

        from collections import Counter
        c = Counter(item for name in detector_coord_names for item in extra_coords[name])
        conflicted = [k for k, v in c.items() if v != 1 and k != 'cycle']

        flat_coordinates = collect_leaves(extra_coords)
        def can_resolve_conflict(c):
            coordinates = flat_coordinates[c]

            if not isinstance(coordinates, list) or len(coordinates) < 2:
                return True

            # check if list of arrays is all equal
            return functools.reduce(lambda x, y: (np.array_equal(x[1], y) and x[0], y), coordinates, (True, coordinates[0]))[0]

        conflicted = [c for c in conflicted if not can_resolve_conflict(c)]

        def clarify_dimensions(dims: List[Dimension], sname: str) -> List[Dimension]:
            return [d if d not in conflicted else d + '-' + sname for d in dims]

        def clarify_coordinate(coordinates: Union[CoordsDict, ndarray], sname: str) -> \
                Union[CoordsDict, ndarray]:
            if not isinstance(coordinates, dict):
                return coordinates

            return {k if k not in conflicted else k + '-' + sname: v for k, v in coordinates.items()}

        dimensions_for_spectra = {k: clarify_dimensions(v, k) for k, v in dimensions_for_spectra.items()}
        extra_coords = {k: clarify_coordinate(v, k) for k, v in extra_coords.items()}
        extra_coords = dict(iter_leaves(extra_coords))

    return extra_coords, dimensions_for_spectra, spectrum_shapes



"""
Phenomenological models and detector simulation for accurate modelling
of ARPES data.

Currently we offer relatively rudimentary detecotor modeling, mostly providing
only nonlinearity and some stubs, but a future release will provide reasonably
accurate modeling of the trapezoidal effect in hemispherical analyzers,
fixed mode artifacts, dust, and more.

Additionally we offer the ability to model the detector response at the
level of individual electron events using a point spread or more complicated
response. This allows the creation of reasonably realistic spectra for testing
new analysis techniques or working on machine learning based approaches that must
be robust to the shortcomings of actual ARPES data.
"""

import numpy as np
import scipy
import scipy.signal as sig

import xarray as xr
from arpes.constants import K_BOLTZMANN_MEV_KELVIN

__all__ = (
    # sampling utilities
    'cloud_to_arr',
    'apply_psf_to_point_cloud',
    'sample_from_distribution',

    # Spectral Function representation
    'SpectralFunction',

    # Composable detector effects, to simulate the real response of
    # an ARPES detector
    'DetectorEffect',

    # implementations of particular spectral functions
    'SpectralFunctionBSSCO', 'SpectralFunctionMFL',
    'SpectralFunctionPhaseCoherent',

    # Particular Detector Models
    'NonlinearDetectorEffect',
    'FixedModeDetectorEffect',
    'DustDetectorEffect',
    'TrapezoidalDetectorEffect',
    'WindowedDetectorEffect',
)


class DetectorEffect:
    """
    Detector effects are callables that map a spectrum into a new
    transformed one. This might be used to imprint the image of a grid,
    dust, or impose detector nonlinearities.
    """
    def __call__(self, spectrum):
        """
        By default, the identity function
        :param spectrum:
        :return:
        """
        return spectrum


class NonlinearDetectorEffect(DetectorEffect):
    """
    Implements power law detector nonlinearities.
    """
    def __init__(self, gamma=None, nonlinearity=None):
        self.gamma = gamma
        self.nonlinearity = nonlinearity

    def __call__(self, spectrum):
        if self.gamma is not None:
            return spectrum ** self.gamma

        raise NotImplementedError(
            'Nonlinearity lookup tables are not yet supported.')


class FixedModeDetectorEffect(DetectorEffect):
    """
    Implements a grid or pore structure of an MCP or field termination mesh.
    Talk to Danny or Sam about getting hyperuniform point cloud distributions to use
    for the pore structure.
    """
    def __init__(self, spacing=None, periodic='hex', detector_efficiency=None):
        if spacing is None:
            spacing = 5 # Five pixel average spacing
        self.spacing = spacing
        self.periodic = periodic
        self._cached_pore_structure = None

    @property
    def detector_imprint(self):
        raise NotImplementedError()

    def __call__(self, spectrum):
        # will fail if we do not have the right size
        return self.detector_imprint * spectrum


class DustDetectorEffect(DetectorEffect):
    """
    TODO, dust.
    """


class TrapezoidalDetectorEffect(DetectorEffect):
    """
    TODO model that phi(pixel) is also a function of binding energy,
    i.e. that the detector has severe aberrations at low photoelectron
    kinetic energy (high retardation ratio).
    """


class WindowedDetectorEffect(DetectorEffect):
    """
    TODO model the finite width of the detector window as recorded on a camera.
    """


def cloud_to_arr(point_cloud, shape):
    """
    Converts a point cloud (list of xy pairs) to an array representation.
    Uses linear interpolation for points that have non-integral coordinates.
    :param point_cloud:
    :param shape:
    :return:
    """
    cloud_as_image = np.zeros(shape)

    for x, y in zip(*point_cloud):
        frac_low_x = 1 - (x - np.floor(x))
        frac_low_y = 1 - (y - np.floor(y))
        shape_x, shape_y = shape
        cloud_as_image[int(np.floor(x)) % shape_x][int(np.floor(y)) % shape_y] += frac_low_x * frac_low_y
        cloud_as_image[(int(np.floor(x)) + 1) % shape_x][int(np.floor(y)) % shape_y] += (1 - frac_low_x) * frac_low_y
        cloud_as_image[int(np.floor(x)) % shape_x][(int(np.floor(y)) + 1) % shape_y] += frac_low_x * (1 - frac_low_y)
        cloud_as_image[(int(np.floor(x)) + 1) % shape_x][(int(np.floor(y)) + 1) % shape_y] += (1 - frac_low_x) * (1 - frac_low_y)

    return cloud_as_image


def apply_psf_to_point_cloud(point_cloud, shape, sigma=None):
    """
    Takes a point cloud and turns it into a spectrum. Finally, smears it by a
    gaussian PSF given through the `sigma` parameter.

    In the future, we should also allow for specifying a particular PSF.
    :param point_cloud:
    :param shape:
    :param sigma:
    :return:
    """
    if sigma is None:
        sigma = (10, 3)
    as_img = cloud_to_arr(point_cloud, shape)

    return scipy.ndimage.gaussian_filter(as_img, sigma=sigma, order=0, mode='reflect')


def sample_from_distribution(distribution, N=5000):
    """
    Given a probability distribution in ND modeled by an array providing the PDF,
    sample individual events coming from this PDF.

    :param distribution:
    :param N:
    :return:
    """
    cdf_rows = np.cumsum(np.sum(distribution.values, axis=1))
    norm_rows = np.cumsum(distribution.values / np.expand_dims(np.sum(distribution.values, axis=1), axis=1), axis=1)

    total = np.sum(distribution.values)

    sample_xs = np.searchsorted(cdf_rows, np.random.random(N,) * total)
    sample_ys_rows = norm_rows[sample_xs, :]

    # take N samples between 0 and 1, which is now the normalized full range of the data
    # and find the index, this effectively samples the index in the array if it were a PDF
    sample_ys = []
    random_ys = np.random.random(N,)
    for random_y, row_y in zip(random_ys, sample_ys_rows):
        sample_ys.append(np.searchsorted(row_y, random_y))

    return (1. * sample_xs + np.random.random(N, )), (1. * np.array(sample_ys) + np.random.random(N, ))


class SpectralFunction:
    """
    Model for a band with self energy.
    """
    def fermi_dirac(self, omega):
        return 1 / (np.exp(omega / (K_BOLTZMANN_MEV_KELVIN * self.temperature)) + 1)

    def __init__(self, k=None, omega=None, temperature=None):
        if temperature is None:
            temperature = 20
        if k is None:
            k = np.linspace(-200, 200, 800)
        elif len(k) == 3:
            k = np.linspace(*k)

        if omega is None:
            omega = np.linspace(-1000, 1000, 2000)
        elif len(omega) == 3:
            omega = np.linspace(*omega)

        self.temperature = temperature
        self.omega = omega
        self.k = k

    def imag_self_energy(self):
        return np.zeros(shape=self.omega.shape,)

    def real_self_energy(self):
        """
        Default to Kramers-Kronig
        """
        return np.imag(sig.hilbert(self.imag_self_energy()))

    def self_energy(self):
        return self.real_self_energy() + 1.j * self.imag_self_energy()

    def bare_band(self):
        return 3 * self.k

    def sampled_spectral_function(self, n_electrons=50000, n_cycles=1, psf=None):
        if psf is None:
            psf = (7, 3)

        spectral = self.measured_spectral_function()
        sampled = [
            apply_psf_to_point_cloud(sample_from_distribution(spectral, N=n_electrons), spectral.values.shape, sigma=psf)
            for _ in range(n_cycles)
        ]

        new_coords = dict(spectral.coords)
        new_coords['cycle'] = np.array(range(n_cycles))
        return xr.DataArray(np.stack(sampled, axis=-1), coords=new_coords, dims=list(spectral.dims) + ['cycle'])

    def measured_spectral_function(self):
        spectral = self.occupied_spectral_function()
        return spectral

    def occupied_spectral_function(self):
        spectral = self.spectral_function()
        spectral.values = spectral.values * np.expand_dims(self.fermi_dirac(self.omega), axis=1)
        return spectral

    def spectral_function(self):
        self_energy = self.self_energy()
        imag_self_energy = np.imag(self_energy)
        real_self_energy = np.real(self_energy)

        bare = self.bare_band()
        full_omegas = np.outer(self.omega, np.ones(shape=bare.shape))

        numerator = np.outer(np.abs(imag_self_energy), np.ones(shape=bare.shape))
        data = numerator / ((full_omegas - np.expand_dims(bare, axis=0) -
                             np.expand_dims(real_self_energy, axis=1)) ** 2 +
                            np.expand_dims(imag_self_energy ** 2, axis=1))
        return xr.DataArray(data, coords={'k': self.k, 'omega': self.omega}, dims=['omega', 'k'])


class SpectralFunctionMFL(SpectralFunction): # pylint: disable=invalid-name
    """
    Implements the Marginal Fermi Liquid spectral function, more or less.
    """

    def __init__(self, k=None, omega=None, temperature=None, a=None, b=None):
        if a is None:
            a = 10

        if b is None:
            b = 1

        super().__init__(k, omega, temperature)

        self.a = a
        self.b = b

    def imag_self_energy(self):
        return np.sqrt((self.a + self.b * self.omega) ** 2 + self.temperature ** 2)


class SpectralFunctionBSSCO(SpectralFunction):
    """
    Implements the spectral function for BSSCO as reported in PhysRevB.57.R11093 and explored in
    `"Collapse of superconductivity in cuprates via ultrafast quenching of phase coherence" <https://arxiv.org/pdf/1707.02305.pdf>`_.
    """

    def __init__(self, k=None, omega=None, temperature=None, delta=None, gamma_s=None, gamma_p=None):
        if delta is None:
            delta = 50
        if gamma_s is None:
            gamma_s = 30
        if gamma_p is None:
            gamma_p = 10

        self.delta = delta
        self.gamma_s = gamma_s
        self.gamma_p = gamma_p
        super().__init__(k, omega, temperature)

    def self_energy(self):
        shape = (len(self.omega), len(self.k))

        g_one = -1.j * self.gamma_s * np.ones(shape=shape)
        bare = self.bare_band()

        full_omegas = np.outer(self.omega, np.ones(shape=bare.shape))

        return g_one + (self.delta ** 2) / (full_omegas + bare + 1.j * self.gamma_p)

    def spectral_function(self):
        self_energy = self.self_energy()
        imag_self_energy = np.imag(self_energy)
        real_self_energy = np.real(self_energy)

        bare = self.bare_band()
        full_omegas = np.outer(self.omega, np.ones(shape=bare.shape))

        numerator = np.abs(imag_self_energy)
        data = numerator / ((full_omegas - np.expand_dims(bare, axis=0) -
                             real_self_energy) ** 2 +
                            imag_self_energy ** 2)
        return xr.DataArray(data, coords={'k': self.k, 'omega': self.omega}, dims=['omega', 'k'])


class SpectralFunctionPhaseCoherent(SpectralFunctionBSSCO):
    def self_energy(self):
        shape = (len(self.omega), len(self.k))

        g_one = -1.j * self.gamma_s * np.ones(shape=shape) * np.sqrt(
            (1 + 0.0005 * np.expand_dims(self.omega, axis=1) ** 2))
        bare = self.bare_band()

        full_omegas = np.outer(self.omega, np.ones(shape=bare.shape))

        self_e = (g_one + (self.delta ** 2) / (full_omegas + bare + 1.j * self.gamma_p))
        imag_self_e = np.imag(self_e)
        real_self_e = np.imag(sig.hilbert(imag_self_e, axis=0))

        return self_e + 3 * (np.random.random(self_e.shape) + np.random.random(self_e.shape) * 1.j)



import warnings

import matplotlib
import matplotlib.animation as manimation
import matplotlib.pyplot as plt
import numpy as np

from arpes.provenance import save_plot_provenance


__all__ = ('plot_subtraction_reference',)

@save_plot_provenance
def plot_subtraction_reference(data, title=None, out=None, norm=None, **kwargs):
    # first we need to sum over dimensions that we do not need

    warnings.warn('Unfinished plot_subtraction_reference')
    return None

    sum_dimensions = {'cycle',}
    sum_dimensions.intersection_update(set(data.dims))

    # summed_data = data.sum(*list(sum_dimensions))

    matplotlib.use("Agg")
    if title is None:
        title = data.S.label.replace('_', ' ')

    FFMpegWriter = manimation.writers['ffmpeg']
    metadata = dict(title=title or 'Test title', artist='Matplotlib')

    writer = FFMpegWriter(fps=15, metadata=metadata)

    fig = plt.figure()
    l, = plt.plot([], [], 'k-o')

    plt.xlim(-5, 5)
    plt.ylim(-5, 5)

    x0, y0 = 0, 0

    with writer.saving(fig, "writer_test.mp4", 100):
        for i in range(100):
            x0 += 0.1 * np.random.randn()
            y0 += 0.1 * np.random.randn()
            l.set_data(x0, y0)
            writer.grab_frame()
